__author__ = "lucaspavanelli"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json
from dotenv import load_dotenv

load_dotenv()
from aixplain.factories import AgentFactory
from aixplain.enums.supplier import Supplier

import pytest

RUN_FILE = "tests/functional/agent/data/agent_test_end2end.json"


def read_data(data_path):
    return json.load(open(data_path, "r"))


@pytest.fixture(scope="module", params=read_data(RUN_FILE))
def run_input_map(request):
    return request.param


def test_end2end(run_input_map):
    for agent in AgentFactory.list()["results"]:
        agent.delete()

    tools = []
    if "model_tools" in run_input_map:
        for tool in run_input_map["model_tools"]:
            for supplier in Supplier:
                if tool["supplier"] is not None and tool["supplier"].lower() in [
                    supplier.value["code"].lower(),
                    supplier.value["name"].lower(),
                ]:
                    tool["supplier"] = supplier
                    break
            tools.append(AgentFactory.create_model_tool(**tool))
    if "pipeline_tools" in run_input_map:
        for tool in run_input_map["pipeline_tools"]:
            tools.append(AgentFactory.create_pipeline_tool(pipeline=tool["pipeline_id"], description=tool["description"]))
    print(f"Creating agent with tools: {tools}")
    agent = AgentFactory.create(name=run_input_map["agent_name"], llm_id=run_input_map["llm_id"], tools=tools)
    print(f"Agent created: {agent.__dict__}")
    print("Running agent")
    response = agent.run(data=run_input_map["query"])
    print(f"Agent response: {response}")
    assert response is not None
    assert response["completed"] is True
    assert response["status"].lower() == "success"
    assert "data" in response
    assert response["data"]["session_id"] is not None
    assert response["data"]["output"] is not None
    print("Deleting agent")
    agent.delete()


def test_list_agents():
    agents = AgentFactory.list()
    assert "results" in agents
    agents_result = agents["results"]
    assert type(agents_result) is list


def test_fail_non_existent_llm():
    with pytest.raises(Exception) as exc_info:
        AgentFactory.create(name="Test Agent", llm_id="non_existent_llm", tools=[])
    assert str(exc_info.value) == "Large Language Model with ID 'non_existent_llm' not found."



__author__ = "mohammedalyafeai"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pytest
from aixplain.enums import License
from aixplain.factories import FileFactory

def test_file_create():
    upload_file = "tests/functional/file_asset/input/test.csv"
    s3_link = FileFactory.create(
        local_path = upload_file,
        tags = ['test1','test2'],
        license = License.MIT,
        is_temp = False
    )
    assert s3_link.startswith("s3")

def test_file_create_temp():
    upload_file = "tests/functional/file_asset/input/test.csv"
    s3_link = FileFactory.create(
        local_path = upload_file,
        tags = ['test1','test2'],
        license = License.MIT,
        is_temp = True
    )
    assert s3_link.startswith("s3")



from typing import List, Union, Type, TYPE_CHECKING, Optional

from aixplain.modules import Model
from aixplain.enums import DataType

from .enums import (
    NodeType,
    FunctionType,
    RouteType,
    Operation,
    AssetType,
)
from .base import (
    Node,
    Link,
    Param,
    InputParam,
    OutputParam,
    TI,
    TO,
    Inputs,
    Outputs,
    Serializable,
)
from .mixins import LinkableMixin, OutputableMixin, RoutableMixin

if TYPE_CHECKING:
    from .pipeline import DesignerPipeline


class AssetNode(Node[TI, TO], LinkableMixin, OutputableMixin):
    """
    Asset node class, this node will be used to fetch the asset from the
    aixplain platform and use it in the pipeline.

    `assetId` is required and will be used to fetch the asset from the
    aixplain platform.

    Input and output parameters will be automatically added based on the
    asset function spec.
    """

    asset_id: Union[Model, str] = None
    function: str = None
    supplier: str = None
    version: str = None
    assetType: AssetType = AssetType.MODEL
    functionType: FunctionType = FunctionType.AI

    type: NodeType = NodeType.ASSET

    def __init__(
        self,
        asset_id: Union[Model, str] = None,
        supplier: str = None,
        version: str = None,
        pipeline: "DesignerPipeline" = None,
    ):
        super().__init__(pipeline=pipeline)
        self.asset_id = asset_id
        self.supplier = supplier
        self.version = version

        if self.asset_id:
            self.populate_asset()

    def populate_asset(self):
        from aixplain.factories.model_factory import ModelFactory

        if isinstance(self.asset_id, str):
            self.asset = ModelFactory.get(self.asset_id)
        elif isinstance(self.asset_id, Model):
            self.asset = self.asset_id
            self.asset_id = self.asset_id.id
        else:
            raise ValueError("assetId should be a string or an Asset instance")

        try:
            self.supplier = self.asset.supplier.value["code"]
        except Exception:
            self.supplier = str(self.asset.supplier)

        self.version = self.asset.version

        if self.function:
            if self.asset.function.value != self.function:
                raise ValueError(
                    f"Function {self.function} is not supported by asset {self.asset_id}"
                )  # noqa
        else:
            self.function = self.asset.function.value
            self._auto_populate_params()

        self._auto_set_params()

    def _auto_populate_params(self):
        from aixplain.enums.function import FunctionInputOutput

        spec = FunctionInputOutput[self.asset.function.value]["spec"]
        for item in spec["params"]:
            self.inputs.create_param(
                code=item["code"],
                data_type=item["dataType"],
                is_required=item["required"],
            )

        for item in spec["output"]:
            self.outputs.create_param(
                code=item["code"],
                data_type=item["dataType"],
            )

    def _auto_set_params(self):
        for k, v in self.asset.additional_info["parameters"].items():
            if isinstance(v, list):
                self.inputs[k] = v[0]
            else:
                self.inputs[k] = v

    def serialize(self) -> dict:
        obj = super().serialize()
        obj["function"] = self.function
        obj["assetId"] = self.asset_id
        obj["supplier"] = self.supplier
        obj["version"] = self.version
        obj["assetType"] = self.assetType
        obj["functionType"] = self.functionType
        obj["type"] = self.type
        return obj


class InputInputs(Inputs):
    pass


class InputOutputs(Outputs):
    input: OutputParam = None

    def __init__(self, node: Node):
        super().__init__(node)
        self.input = self.create_param("input")


class Input(Node[InputInputs, InputOutputs], LinkableMixin, RoutableMixin):
    """
    Input node class, this node will be used to input the data to the
    pipeline.

    Input nodes has only one output parameter called `input`.

    `data` is a special convenient parameter that will be uploaded to the
    aixplain platform and the link will be passed as the input to the node.
    """

    data_types: Optional[List[DataType]] = None
    data: Optional[str] = None
    type: NodeType = NodeType.INPUT
    inputs_class: Type[TI] = InputInputs
    outputs_class: Type[TO] = InputOutputs

    def __init__(
        self,
        data: Optional[str] = None,
        data_types: Optional[List[DataType]] = None,
        pipeline: "DesignerPipeline" = None,
    ):
        from aixplain.factories.file_factory import FileFactory

        super().__init__(pipeline=pipeline)
        self.data_types = data_types or []
        self.data = data

        if self.data:
            self.data = FileFactory.to_link(self.data, is_temp=True)

    def serialize(self) -> dict:
        obj = super().serialize()
        obj["data"] = self.data
        obj["dataType"] = self.data_types
        return obj


class OutputInputs(Inputs):
    output: InputParam = None

    def __init__(self, node: Node):
        super().__init__(node)
        self.output = self.create_param("output")


class OutputOutputs(Outputs):
    pass


class Output(Node[OutputInputs, OutputOutputs]):
    """
    Output node class, this node will be used to output the result of the
    pipeline.

    Output nodes has only one input parameter called `output`.
    """

    data_types: Optional[List[DataType]] = None
    type: NodeType = NodeType.OUTPUT
    inputs_class: Type[TI] = OutputInputs
    outputs_class: Type[TO] = OutputOutputs

    def __init__(
        self,
        data_types: Optional[List[DataType]] = None,
        pipeline: "DesignerPipeline" = None,
    ):
        super().__init__(pipeline=pipeline)
        self.data_types = data_types or []

    def serialize(self) -> dict:
        obj = super().serialize()
        obj["dataType"] = self.data_types
        return obj


class Script(Node[TI, TO], LinkableMixin, OutputableMixin):
    """
    Script node class, this node will be used to run a script on the input
    data.

    `script_path` is a special convenient parameter that will be uploaded to
    the aixplain platform and the link will be passed as the input to the node.
    """

    fileId: Optional[str] = None
    script_path: Optional[str] = None
    type: NodeType = NodeType.SCRIPT

    def __init__(
        self,
        pipeline: "DesignerPipeline" = None,
        script_path: Optional[str] = None,
        fileId: Optional[str] = None,
    ):
        from aixplain.factories.script_factory import ScriptFactory

        super().__init__(pipeline=pipeline)

        assert script_path or fileId, "script_path or fileId is required"

        if not fileId:
            self.fileId = ScriptFactory.upload_script(script_path)
        else:
            self.fileId = fileId

    def serialize(self) -> dict:
        obj = super().serialize()
        obj["fileId"] = self.fileId
        return obj


class Route(Serializable):
    """
    Route class, this class will be used to route the input data to different
    nodes based on the input data type.
    """

    value: DataType
    path: List[Union[Node, int]]
    operation: Operation
    type: RouteType

    def __init__(
        self,
        value: DataType,
        path: List[Union[Node, int]],
        operation: Operation,
        type: RouteType,
    ):
        """
        Post init method to convert the nodes to node numbers if they are
        nodes.
        """
        self.value = value
        self.path = path
        self.operation = operation
        self.type = type

        if not self.path:
            raise ValueError("Path is not valid, should be a list of nodes")

        # convert nodes to node numbers if they are nodes
        self.path = [
            node.number if isinstance(node, Node) else node
            for node in self.path
        ]

    def serialize(self) -> dict:
        return {
            "value": self.value,
            "path": self.path,
            "operation": self.operation,
            "type": self.type,
        }


class RouterInputs(Inputs):
    input: InputParam = None

    def __init__(self, node: Node):
        super().__init__(node)
        self.input = self.create_param("input")


class RouterOutputs(Outputs):
    input: OutputParam = None

    def __init__(self, node: Node):
        super().__init__(node)
        self.input = self.create_param("input")


class Router(Node[RouterInputs, RouterOutputs], LinkableMixin):
    """
    Router node class, this node will be used to route the input data to
    different nodes based on the input data type.
    """

    routes: List[Route]
    type: NodeType = NodeType.ROUTER
    inputs_class: Type[TI] = RouterInputs
    outputs_class: Type[TO] = RouterOutputs

    def __init__(
        self, routes: List[Route], pipeline: "DesignerPipeline" = None
    ):
        super().__init__(pipeline=pipeline)
        self.routes = routes

    def serialize(self) -> dict:
        obj = super().serialize()
        obj["routes"] = [route.serialize() for route in self.routes]
        return obj


class DecisionInputs(Inputs):
    comparison: InputParam = None
    passthrough: InputParam = None

    def __init__(self, node: Node):
        super().__init__(node)
        self.comparison = self.create_param("comparison")
        self.passthrough = self.create_param("passthrough")


class DecisionOutputs(Outputs):
    input: OutputParam = None

    def __init__(self, node: Node):
        super().__init__(node)
        self.input = self.create_param("input")


class Decision(Node[DecisionInputs, DecisionOutputs], LinkableMixin):
    """
    Decision node class, this node will be used to make decisions based on
    the input data.
    """

    routes: List[Route]
    type: NodeType = NodeType.DECISION
    inputs_class: Type[TI] = DecisionInputs
    outputs_class: Type[TO] = DecisionOutputs

    def __init__(
        self, routes: List[Route], pipeline: "DesignerPipeline" = None
    ):
        super().__init__(pipeline=pipeline)
        self.routes = routes

    def link(
        self,
        to_node: Node,
        from_param: Union[str, Param],
        to_param: Union[str, Param],
    ) -> Link:
        link = super().link(to_node, from_param, to_param)
        self.outputs.input.data_type = self.inputs.passthrough.data_type
        return link

    def serialize(self) -> dict:
        obj = super().serialize()
        obj["routes"] = [route.serialize() for route in self.routes]
        return obj


class BaseSegmentor(AssetNode[TI, TO]):
    """
    Segmentor node class, this node will be used to segment the input data
    into smaller fragments for much easier and efficient processing.
    """

    type: NodeType = NodeType.SEGMENTOR
    functionType: FunctionType = FunctionType.SEGMENTOR


class SegmentorInputs(Inputs):
    pass


class SegmentorOutputs(Outputs):
    audio: OutputParam = None

    def __init__(self, node: Node):
        super().__init__(node)
        self.audio = self.create_param("audio")


class BareSegmentor(BaseSegmentor[SegmentorInputs, SegmentorOutputs]):
    """
    Segmentor node class, this node will be used to segment the input data
    into smaller fragments for much easier and efficient processing.
    """

    type: NodeType = NodeType.SEGMENTOR
    functionType: FunctionType = FunctionType.SEGMENTOR
    inputs_class: Type[TI] = SegmentorInputs
    outputs_class: Type[TO] = SegmentorOutputs


class BaseReconstructor(AssetNode[TI, TO]):
    """
    Reconstructor node class, this node will be used to reconstruct the
    output of the segmented lines of execution.
    """

    type: NodeType = NodeType.RECONSTRUCTOR
    functionType: FunctionType = FunctionType.RECONSTRUCTOR


class ReconstructorInputs(Inputs):
    data: InputParam = None

    def __init__(self, node: Node):
        super().__init__(node)
        self.data = self.create_param("data")


class ReconstructorOutputs(Outputs):
    data: OutputParam = None

    def __init__(self, node: Node):
        super().__init__(node)
        self.data = self.create_param("data")


class BareReconstructor(
    BaseReconstructor[ReconstructorInputs, ReconstructorOutputs]
):
    """
    Reconstructor node class, this node will be used to reconstruct the
    output of the segmented lines of execution.
    """

    type: NodeType = NodeType.RECONSTRUCTOR
    functionType: FunctionType = FunctionType.RECONSTRUCTOR
    inputs_class: Type[TI] = ReconstructorInputs
    outputs_class: Type[TO] = ReconstructorOutputs



__author__ = "aiXplain"

"""
Copyright 2024 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Lucas Pavanelli and Thiago Castro Ferreira
Date: May 16th 2024
Description:
    Agentification Class
"""
from typing import Optional, Union, Text

from aixplain.enums.function import Function
from aixplain.enums.supplier import Supplier
from aixplain.modules.agent.tool import Tool
from aixplain.modules.model import Model


class ModelTool(Tool):
    """Specialized software or resource designed to assist the AI in executing specific tasks or functions based on user commands.

    Attributes:
        function (Function): task that the tool performs
        supplier (Optional[Union[Dict, Text, Supplier, int]], optional): Preferred supplier to perform the task. Defaults to None.
    """

    def __init__(
        self,
        function: Optional[Function] = None,
        supplier: Optional[Supplier] = None,
        model: Optional[Union[Text, Model]] = None,
        **additional_info,
    ) -> None:
        """Specialized software or resource designed to assist the AI in executing specific tasks or functions based on user commands.

        Args:
            function (Optional[Function], optional): task that the tool performs. Defaults to None.
            supplier (Optional[Supplier], optional): Preferred supplier to perform the task. Defaults to None.. Defaults to None.
            model (Optional[Union[Text, Model]], optional): Model function. Defaults to None.
        """
        assert (
            function is not None or model is not None
        ), "Agent Creation Error: Either function or model must be provided when instantiating a tool."
        super().__init__("", "", **additional_info)
        if function is not None:
            if isinstance(function, str):
                function = Function(function)

        try:
            if isinstance(supplier, dict):
                supplier = Supplier(supplier)
        except Exception:
            supplier = None

        if model is not None:
            if isinstance(model, Text) is True:
                self.model = model
                model = self.validate()
            function = model.function
            if isinstance(model.supplier, Supplier):
                supplier = model.supplier
            model = model.id
        self.supplier = supplier
        self.model = model
        self.function = function

    def validate(self) -> Model:
        from aixplain.factories.model_factory import ModelFactory

        try:
            model = None
            if self.model is not None:
                model = ModelFactory.get(self.model)
            return model
        except Exception:
            raise Exception(f"Model Tool Unavailable. Make sure Model '{self.model}' exists or you have access to it.")



__author__ = "thiagocastroferreira"

import json
import logging
import os
import pandas as pd
import tarfile

from aixplain.enums.data_type import DataType
from aixplain.enums.file_type import FileType
from aixplain.enums.storage_type import StorageType
from aixplain.modules.content_interval import (
    ContentInterval,
    AudioContentInterval,
    ImageContentInterval,
    TextContentInterval,
    VideoContentInterval,
)
from aixplain.modules.file import File
from aixplain.modules.metadata import MetaData
from aixplain.utils.file_utils import upload_data
from pathlib import Path
from tqdm import tqdm
from typing import Any, Dict, List, Text, Tuple


def compress_folder(folder_path: str):
    with tarfile.open(folder_path + ".tgz", "w:gz") as tar:
        for name in os.listdir(folder_path):
            tar.add(os.path.join(folder_path, name))
    return folder_path + ".tgz"


def process_interval(interval: Any, storage_type: StorageType, interval_folder: Text) -> List[Dict]:
    """Process text files

    Args:
        intervals (Any): content intervals to process the content
        storage_type (StorageType): type of storage: URL, local path or textual content

    Returns:
        List[Dict]: content interval
    """
    if storage_type == StorageType.FILE:
        # Check the size of file and assert a limit of 50 MB
        assert (
            os.path.getsize(interval.content) <= 25000000
        ), f'Data Asset Onboarding Error: Local text file "{interval}" exceeds the size limit of 25 MB.'
        fname = os.path.basename(interval)
        new_path = os.path.join(audio_folder, fname)
        if os.path.exists(new_path) is False:
            shutil.copy2(audio_path, new_path)
    return [interval.__dict__ for interval in intervals]


def validate_format(index: int, interval: Dict, metadata: MetaData) -> ContentInterval:
    """Validate the interval format

    Args:
        index (int): row index
        interval (Dict): interval to be validated
        metadata (MetaData): metadata

    Returns:
        ContentInterval: _description_
    """
    if metadata.dtype == DataType.AUDIO_INTERVAL:
        try:
            if isinstance(interval, list):
                interval = [AudioContentInterval(**interval_) for interval_ in interval]
            else:
                interval = [AudioContentInterval(**interval)]
        except Exception as e:
            message = f'Data Asset Onboarding Error: Audio Interval in row {index} of Column "{metadata.name}" is not following the format. Check the "AudioContentInterval" class for the correct format.'
            logging.exception(message)
            raise Exception(message)
    elif metadata.dtype == DataType.IMAGE_INTERVAL:
        try:
            if isinstance(interval, list):
                interval = [ImageContentInterval(**interval_) for interval_ in interval]
            else:
                interval = [ImageContentInterval(**interval)]
        except Exception as e:
            message = f'Data Asset Onboarding Error: Image Interval in row {index} of Column "{metadata.name}" is not following the format. Check the "ImageContentInterval" class for the correct format.'
            logging.exception(message)
            raise Exception(message)
    elif metadata.dtype == DataType.TEXT_INTERVAL:
        try:
            if isinstance(interval, list):
                interval = [TextContentInterval(**interval_) for interval_ in interval]
            else:
                interval = [TextContentInterval(**interval)]
        except Exception as e:
            message = f'Data Asset Onboarding Error: Text Interval in row {index} of Column "{metadata.name}" is not following the format. Check the "TextContentInterval" class for the correct format.'
            logging.exception(message)
            raise Exception(message)
    elif metadata.dtype == DataType.VIDEO_INTERVAL:
        try:
            if isinstance(interval, list):
                interval = [VideoContentInterval(**interval_) for interval_ in interval]
            else:
                interval = [VideoContentInterval(**interval)]
        except Exception as e:
            message = f'Data Asset Onboarding Error: Video Interval in row {index} of Column "{metadata.name}" is not following the format. Check the "VideoContentInterval" class for the correct format.'
            logging.exception(message)
            raise Exception(message)
    return interval


def run(metadata: MetaData, paths: List, folder: Path, batch_size: int = 1000) -> Tuple[List[File], int, int]:
    """Process a list of local interval files, compress and upload them to pre-signed URLs in S3

    Explanation:
        Each interval on "paths" is processed. If the interval content is in a public link or local file, it will be downloaded and added to an index CSV file.
        The intervals are processed in batches such that at each "batch_size" texts, the index CSV file is uploaded into a pre-signed URL in s3 and reset.

    Args:
        metadata (MetaData): meta data of the asset
        paths (List): list of paths to local files
        folder (Path): local folder to save compressed files before upload them to s3.

    Returns:
        Tuple[List[File], int, int]: list of s3 links, data colum index and number of rows
    """
    logging.debug(f'Data Asset Onboarding: Processing "{metadata.name}".')
    interval_folder = Path(".")
    if metadata.storage_type in [StorageType.FILE, StorageType.TEXT]:
        interval_folder = Path(os.path.join(folder, "data"))
        interval_folder.mkdir(exist_ok=True)

    idx = 0
    data_column_idx = -1
    files, batch = [], []
    for i in tqdm(range(len(paths)), desc=f' Data "{metadata.name}" onboarding progress', position=1, leave=False):
        path = paths[i]
        try:
            dataframe = pd.read_csv(path)
        except Exception as e:
            message = f'Data Asset Onboarding Error: Local file "{path}" not found.'
            logging.exception(message)
            raise Exception(message)

        # process intervals
        for j in tqdm(range(len(dataframe)), desc=" File onboarding progress", position=2, leave=False):
            row = dataframe.iloc[j]
            try:
                interval = row[metadata.name]
            except Exception as e:
                message = f'Data Asset Onboarding Error: Column "{metadata.name}" not found in the local file {path}.'
                logging.exception(message)
                raise Exception(message)

            # interval = validate_format(index=j, interval=interval, metadata=metadata)

            try:
                interval = process_interval(interval, metadata.storage_type)
                batch.append(interval)
            except Exception as e:
                logging.exception(e)
                raise Exception(e)

            idx += 1
            if ((idx) % batch_size) == 0:
                batch_index = str(len(files) + 1).zfill(8)
                file_name = f"{folder}/{metadata.name}-{batch_index}.csv.gz"

                df = pd.DataFrame({metadata.name: batch})
                start, end = idx - len(batch), idx
                df["@INDEX"] = range(start, end)
                df.to_csv(file_name, compression="gzip", index=False)
                s3_link = upload_data(file_name, content_type="text/csv", content_encoding="gzip")
                files.append(File(path=s3_link, extension=FileType.CSV, compression="gzip"))
                # get data column index
                data_column_idx = df.columns.to_list().index(metadata.name)
                batch = []

    if len(batch) > 0:
        batch_index = str(len(files) + 1).zfill(8)
        file_name = f"{folder}/{metadata.name}-{batch_index}.csv.gz"

        df = pd.DataFrame({metadata.name: batch})
        start, end = idx - len(batch), idx
        df["@INDEX"] = range(start, end)
        df.to_csv(file_name, compression="gzip", index=False)
        s3_link = upload_data(file_name, content_type="text/csv", content_encoding="gzip")
        files.append(File(path=s3_link, extension=FileType.CSV, compression="gzip"))
        # get data column index
        data_column_idx = df.columns.to_list().index(metadata.name)
        batch = []
    return files, data_column_idx, idx



__author__ = "thiagocastroferreira"

"""
Copyright 2024 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Duraikrishna Selvaraju, Thiago Castro Ferreira, Shreyas Sharma and Lucas Pavanelli
Date: February 21st 2024
Description:
    FinetuneCost Class
"""

from aixplain.enums.asset_status import AssetStatus
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from typing import Optional, Text

@dataclass_json
@dataclass
class FinetuneStatus(object):
    status: "AssetStatus"
    model_status: "AssetStatus"
    epoch: Optional[float] = None
    training_loss: Optional[float] = None
    validation_loss: Optional[float] = None



__author__ = "thiagocastroferreira"

import pytest

from aixplain.enums import Function
from aixplain.factories import ModelFactory
from aixplain.modules import LLM


@pytest.mark.parametrize("llm_model", ["Groq Llama 3 70B", "Chat GPT 3.5", "GPT-4o", "GPT 4 (32k)"])
def test_llm_run(llm_model):
    """Testing LLMs with history context"""
    model = ModelFactory.list(query=llm_model, function=Function.TEXT_GENERATION)["results"][0]

    assert isinstance(model, LLM)

    response = model.run(
        data="What is my name?",
        history=[{"role": "user", "content": "Hello! My name is Thiago."}, {"role": "assistant", "content": "Hello!"}],
    )
    assert response["status"] == "SUCCESS"
    assert "thiago" in response["data"].lower()



"""
aiXplain SDK Library.
---

aiXplain SDK enables python programmers to add AI functions
to their software.

Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from .asset import Asset
from .corpus import Corpus
from .data import Data
from .dataset import Dataset
from .file import File
from .metadata import MetaData
from .metric import Metric
from .model import Model
from .model.llm_model import LLM
from .pipeline import Pipeline
from .finetune import Finetune, FinetuneCost
from .finetune.status import FinetuneStatus
from .benchmark import Benchmark
from .benchmark_job import BenchmarkJob
from .agent import Agent
from .agent.tool import Tool



import uuid
import pandas as pd
import json
from dotenv import load_dotenv
import time

load_dotenv()
from aixplain.factories import ModelFactory, DatasetFactory, MetricFactory, BenchmarkFactory
from aixplain.modules.benchmark import Benchmark
from aixplain.modules.benchmark_job import BenchmarkJob
from pathlib import Path

import pytest

import logging

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

TIMEOUT = 60 * 30
RUN_FILE = str(Path(r"tests/functional/benchmark/data/benchmark_test_run_data.json"))
MODULE_FILE = str(Path(r"tests/functional/benchmark/data/benchmark_module_test_data.json"))


def read_data(data_path):
    return json.load(open(data_path, "r"))


@pytest.fixture(scope="module", params=read_data(RUN_FILE))
def run_input_map(request):
    return request.param


@pytest.fixture(scope="module", params=read_data(MODULE_FILE))
def module_input_map(request):
    return request.param

def is_job_finshed(benchmark_job):
    time_taken = 0
    sleep_time = 15
    timeout = 10 * 60
    while True:
        if time_taken > timeout:
            break
        job_status = benchmark_job.check_status()
        if job_status == "in_progress":
            time.sleep(sleep_time)
            time_taken += sleep_time
        elif job_status == "completed":
            return True
        else:
            break
    return False

def assert_correct_results(benchmark_job):
    df = benchmark_job.download_results_as_csv(return_dataframe=True)
    assert type(df) is pd.DataFrame, "Couldn't download CSV"
    model_success_rate = (sum(df["Model_success"])*100)/len(df.index)
    assert model_success_rate >  80 , f"Low model success rate ({model_success_rate})"
    metric_name = "BLEU by sacrebleu"
    mean_score = df[metric_name].mean()
    assert mean_score != 0 , f"Zero Mean Score - Please check metric ({metric_name})"



def test_create_and_run(run_input_map):
    model_list = [ModelFactory.get(model_id) for model_id in run_input_map["model_ids"]]
    dataset_list = [DatasetFactory.list(query=dataset_name)["results"][0] for dataset_name in run_input_map["dataset_names"]]
    metric_list = [MetricFactory.get(metric_id) for metric_id in run_input_map["metric_ids"]]
    benchmark = BenchmarkFactory.create(f"SDK Benchmark Test {uuid.uuid4()}", dataset_list, model_list, metric_list)
    assert type(benchmark) is Benchmark, "Couldn't create benchmark"
    benchmark_job = benchmark.start()
    assert type(benchmark_job) is BenchmarkJob, "Couldn't start job"
    assert is_job_finshed(benchmark_job), "Job did not finish in time"
    assert_correct_results(benchmark_job)





# def test_module(module_input_map):
#     benchmark = BenchmarkFactory.get(module_input_map["benchmark_id"])
#     assert benchmark.id == module_input_map["benchmark_id"]
#     benchmark_job = benchmark.job_list[0]
#     assert benchmark_job.benchmark_id == module_input_map["benchmark_id"]
#     job_status = benchmark_job.check_status()
#     assert job_status in ["in_progress", "completed"]
#     df = benchmark_job.download_results_as_csv(return_dataframe=True)
#     assert type(df) is pd.DataFrame



import aixplain.utils.config as config
from aixplain.modules.wallet import Wallet
from aixplain.utils.file_utils import _request_with_retry
import logging
from typing import Text


class WalletFactory:
    aixplain_key = config.AIXPLAIN_API_KEY
    backend_url = config.BACKEND_URL

    @classmethod
    def get(cls, api_key: Text = config.TEAM_API_KEY) -> Wallet:     
        """Get wallet information"""
        try:
            resp = None
            url = f"{cls.backend_url}/sdk/billing/wallet"
            headers = {"Authorization": f"Token {api_key}", "Content-Type": "application/json"}
            logging.info(f"Start fetching billing information from - {url} - {headers}")
            headers = {"Content-Type": "application/json", "x-api-key": api_key}
            r = _request_with_retry("get", url, headers=headers)
            resp = r.json()
            total_balance = float(resp.get("totalBalance", 0.0))
            reserved_balance = float(resp.get("reservedBalance", 0.0))
        
            return Wallet(total_balance=total_balance, reserved_balance=reserved_balance)
        except Exception as e:
            raise Exception(f"Failed to get the wallet credit information. Error: {str(e)}")



__author__ = "thiagocastroferreira"

import argparse
import json


def main(transcripts, speakers, output_file):
    # get the speech recognition json
    transcripts = json.load(open(transcripts))
    # get the speaker diarization json
    speakers = json.load(open(speakers))

    # build the response
    response = []
    for i, transcript in enumerate(transcripts):
        merge = {
            "transcript": transcript["attributes"]["data"],
            "speaker": speakers[i]["attributes"]["data"]["data"],
        }
        response.append(
            {
                "index": i,
                "success": True,
                "input_type": "text",
                "is_url": transcript["is_url"],
                "details": {},
                "input_segment_info": transcript["input_segment_info"],
                "attributes": {"data": merge, "input": merge},
            }
        )

    # save the response, based on the intermediate representation format, in the output_file
    with open(output_file, "w") as f:
        json.dump(response, f)


if __name__ == "__main__":
    # Create the parser
    parser = argparse.ArgumentParser()
    # Add arguments
    parser.add_argument("--transcripts", type=str, required=True)
    parser.add_argument("--speakers", type=str, required=True)
    parser.add_argument("--output_file", type=str, required=True)
    # Parse the argument
    args = parser.parse_args()

    transcripts = args.transcripts
    speakers = args.speakers
    output_file = args.output_file

    main(transcripts, speakers, output_file)



__author__ = "aiXplain"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Duraikrishna Selvaraju, Thiago Castro Ferreira, Shreyas Sharma and Lucas Pavanelli
Date: September 1st 2022
Description:
    Model Factory Class
"""
from typing import Dict, List, Optional, Text, Tuple, Union
import json
import logging
from aixplain.modules.model import Model
from aixplain.modules.model.llm_model import LLM
from aixplain.enums import Function, Language, OwnershipType, Supplier, SortBy, SortOrder
from aixplain.utils import config
from aixplain.utils.file_utils import _request_with_retry
from urllib.parse import urljoin
from warnings import warn
from aixplain.enums.function import FunctionInputOutput
from datetime import datetime


class ModelFactory:
    """A static class for creating and exploring Model Objects.

    Attributes:
        backend_url (str): The URL for the backend.
    """

    aixplain_key = config.AIXPLAIN_API_KEY
    backend_url = config.BACKEND_URL

    @classmethod
    def _create_model_from_response(cls, response: Dict) -> Model:
        """Converts response Json to 'Model' object

        Args:
            response (Dict): Json from API

        Returns:
            Model: Coverted 'Model' object
        """
        if "api_key" not in response:
            response["api_key"] = config.TEAM_API_KEY

        parameters = {}
        if "params" in response:
            for param in response["params"]:
                if "language" in param["name"]:
                    parameters[param["name"]] = [w["value"] for w in param["values"]]

        function = Function(response["function"]["id"])
        ModelClass = Model
        if function == Function.TEXT_GENERATION:
            ModelClass = LLM

        created_at = None
        if "createdAt" in response and response["createdAt"]:
            created_at = datetime.fromisoformat(response["createdAt"].replace("Z", "+00:00"))
        function_id = response["function"]["id"]
        function = Function(function_id)
        function_io = FunctionInputOutput.get(function_id, None)
        input_params = {param["code"]: param for param in function_io["spec"]["params"]}
        output_params = {param["code"]: param for param in function_io["spec"]["output"]}

        return ModelClass(
            response["id"],
            response["name"],
            supplier=response["supplier"],
            api_key=response["api_key"],
            cost=response["pricing"],
            function=function,
            created_at=created_at,
            parameters=parameters,
            input_params=input_params,
            output_params=output_params,
            is_subscribed=True if "subscription" in response else False,
            version=response["version"]["id"],
        )

    @classmethod
    def get(cls, model_id: Text, api_key: Optional[Text] = None) -> Model:
        """Create a 'Model' object from model id

        Args:
            model_id (Text): Model ID of required model.
            api_key (Optional[Text], optional): Model API key. Defaults to None.

        Returns:
            Model: Created 'Model' object
        """
        resp = None
        try:
            url = urljoin(cls.backend_url, f"sdk/models/{model_id}")
            if cls.aixplain_key != "":
                headers = {"x-aixplain-key": f"{cls.aixplain_key}", "Content-Type": "application/json"}
            else:
                headers = {"Authorization": f"Token {config.TEAM_API_KEY}", "Content-Type": "application/json"}
            logging.info(f"Start service for GET Model  - {url} - {headers}")
            r = _request_with_retry("get", url, headers=headers)
            resp = r.json()
            # set api key
            resp["api_key"] = config.TEAM_API_KEY
            if api_key is not None:
                resp["api_key"] = api_key
            model = cls._create_model_from_response(resp)
            logging.info(f"Model Creation: Model {model_id} instantiated.")
            return model
        except Exception:
            if resp is not None and "statusCode" in resp:
                status_code = resp["statusCode"]
                message = resp["message"]
                message = f"Model Creation: Status {status_code} - {message}"
            else:
                message = "Model Creation: Unspecified Error"
            logging.error(message)
            raise Exception(f"{message}")

    @classmethod
    def create_asset_from_id(cls, model_id: Text) -> Model:
        warn(
            'This method will be deprecated in the next versions of the SDK. Use "get" instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        return cls.get(model_id)

    @classmethod
    def _get_assets_from_page(
        cls,
        query,
        page_number: int,
        page_size: int,
        function: Function,
        suppliers: Union[Supplier, List[Supplier]],
        source_languages: Union[Language, List[Language]],
        target_languages: Union[Language, List[Language]],
        is_finetunable: bool = None,
        ownership: Optional[Tuple[OwnershipType, List[OwnershipType]]] = None,
        sort_by: Optional[SortBy] = None,
        sort_order: SortOrder = SortOrder.ASCENDING,
    ) -> List[Model]:
        try:
            url = urljoin(cls.backend_url, "sdk/models/paginate")
            filter_params = {"q": query, "pageNumber": page_number, "pageSize": page_size}
            if is_finetunable is not None:
                filter_params["isFineTunable"] = is_finetunable
            if function is not None:
                filter_params["functions"] = [function.value]
            if suppliers is not None:
                if isinstance(suppliers, Supplier) is True:
                    suppliers = [suppliers]
                filter_params["suppliers"] = [supplier.value["id"] for supplier in suppliers]
            if ownership is not None:
                if isinstance(ownership, OwnershipType) is True:
                    ownership = [ownership]
                filter_params["ownership"] = [ownership_.value for ownership_ in ownership]

            lang_filter_params = []
            if source_languages is not None:
                if isinstance(source_languages, Language):
                    source_languages = [source_languages]
                if function == Function.TRANSLATION:
                    lang_filter_params.append({"code": "sourcelanguage", "value": source_languages[0].value["language"]})
                else:
                    lang_filter_params.append({"code": "language", "value": source_languages[0].value["language"]})
                    if source_languages[0].value["dialect"] != "":
                        lang_filter_params.append({"code": "dialect", "value": source_languages[0].value["dialect"]})
            if target_languages is not None:
                if isinstance(target_languages, Language):
                    target_languages = [target_languages]
                if function == Function.TRANSLATION:
                    code = "targetlanguage"
                    lang_filter_params.append({"code": code, "value": target_languages[0].value["language"]})
            if sort_by is not None:
                filter_params["sort"] = [{"dir": sort_order.value, "field": sort_by.value}]
            if len(lang_filter_params) != 0:
                filter_params["ioFilter"] = lang_filter_params
            if cls.aixplain_key != "":
                headers = {"x-aixplain-key": f"{cls.aixplain_key}", "Content-Type": "application/json"}
            else:
                headers = {"Authorization": f"Token {config.TEAM_API_KEY}", "Content-Type": "application/json"}

            logging.info(f"Start service for POST Models Paginate - {url} - {headers} - {json.dumps(filter_params)}")
            r = _request_with_retry("post", url, headers=headers, json=filter_params)
            resp = r.json()
            logging.info(f"Listing Models: Status of getting Models on Page {page_number}: {r.status_code}")
            all_models = resp["items"]
            model_list = [cls._create_model_from_response(model_info_json) for model_info_json in all_models]
            return model_list, resp["total"]
        except Exception as e:
            error_message = f"Listing Models: Error in getting Models on Page {page_number}: {e}"
            logging.error(error_message, exc_info=True)
            return []

    @classmethod
    def list(
        cls,
        query: Optional[Text] = "",
        function: Optional[Function] = None,
        suppliers: Optional[Union[Supplier, List[Supplier]]] = None,
        source_languages: Optional[Union[Language, List[Language]]] = None,
        target_languages: Optional[Union[Language, List[Language]]] = None,
        is_finetunable: Optional[bool] = None,
        ownership: Optional[Tuple[OwnershipType, List[OwnershipType]]] = None,
        sort_by: Optional[SortBy] = None,
        sort_order: SortOrder = SortOrder.ASCENDING,
        page_number: int = 0,
        page_size: int = 20,
    ) -> List[Model]:
        """Gets the first k given models based on the provided task and language filters

        Args:
            function (Optional[Function], optional): function filter. Defaults to None.
            source_languages (Optional[Union[Language, List[Language]]], optional): language filter of input data. Defaults to None.
            target_languages (Optional[Union[Language, List[Language]]], optional): language filter of output data. Defaults to None.
            is_finetunable (Optional[bool], optional): can be finetuned or not. Defaults to None.
            ownership (Optional[Tuple[OwnershipType, List[OwnershipType]]], optional): Ownership filters (e.g. SUBSCRIBED, OWNER). Defaults to None.
            sort_by (Optional[SortBy], optional): sort the retrived models by a specific attribute,
            page_number (int, optional): page number. Defaults to 0.
            page_size (int, optional): page size. Defaults to 20.

        Returns:
            List[Model]: List of models based on given filters
        """
        try:
            models, total = cls._get_assets_from_page(
                query,
                page_number,
                page_size,
                function,
                suppliers,
                source_languages,
                target_languages,
                is_finetunable,
                ownership,
                sort_by,
                sort_order,
            )
            return {
                "results": models,
                "page_total": min(page_size, len(models)),
                "page_number": page_number,
                "total": total,
            }
        except Exception as e:
            error_message = f"Listing Models: Error in Listing Models : {e}"
            logging.error(error_message, exc_info=True)
            raise Exception(error_message)

    @classmethod
    def list_host_machines(cls, api_key: Optional[Text] = None) -> List[Dict]:
        """Lists available hosting machines for model.

        Args:
            api_key (Text, optional): Team API key. Defaults to None.

        Returns:
            List[Dict]: List of dictionaries containing information about
            each hosting machine.
        """
        machines_url = urljoin(config.BACKEND_URL, "sdk/hosting-machines")
        logging.debug(f"URL: {machines_url}")
        if api_key:
            headers = {"x-api-key": f"{api_key}", "Content-Type": "application/json"}
        else:
            headers = {"x-api-key": f"{config.TEAM_API_KEY}", "Content-Type": "application/json"}
        response = _request_with_retry("get", machines_url, headers=headers)
        response_dicts = json.loads(response.text)
        for dictionary in response_dicts:
            del dictionary["id"]
        return response_dicts

    @classmethod
    def list_gpus(cls, api_key: Optional[Text] = None) -> List[List[Text]]:
        """List GPU names on which you can host your language model.

        Args:
            api_key (Text, optional): Team API key. Defaults to None.

        Returns:
            List[List[Text]]: List of all available GPUs and their prices.
        """
        gpu_url = urljoin(config.BACKEND_URL, "sdk/model-onboarding/gpus")
        if api_key:
            headers = {"Authorization": f"Token {api_key}", "Content-Type": "application/json"}
        else:
            headers = {"Authorization": f"Token {config.TEAM_API_KEY}", "Content-Type": "application/json"}
        response = _request_with_retry("get", gpu_url, headers=headers)
        response_list = json.loads(response.text)
        return response_list

    @classmethod
    def list_functions(cls, verbose: Optional[bool] = False, api_key: Optional[Text] = None) -> List[Dict]:
        """Lists supported model functions on platform.

        Args:
            verbose (Boolean, optional): Set to True if a detailed response
                is desired; is otherwise False by default.
            api_key (Text, optional): Team API key. Defaults to None.

        Returns:
            List[Dict]: List of dictionaries containing information about
            each supported function.
        """
        functions_url = urljoin(config.BACKEND_URL, "sdk/functions")
        logging.debug(f"URL: {functions_url}")
        if api_key:
            headers = {"x-api-key": f"{api_key}", "Content-Type": "application/json"}
        else:
            headers = {"x-api-key": f"{config.TEAM_API_KEY}", "Content-Type": "application/json"}
        response = _request_with_retry("get", functions_url, headers=headers)
        response_dict = json.loads(response.text)
        if verbose:
            return response_dict
        del response_dict["results"]
        function_list = response_dict["items"]
        for function_dict in function_list:
            del function_dict["output"]
            del function_dict["params"]
            del function_dict["id"]
        return response_dict

    # Will add "always_on" and "is_async" when we support them.
    # def create_asset_repo(cls, name: Text, hosting_machine: Text, version: Text,
    #                       description: Text, function: Text, is_async: bool,
    #                       source_language: Text, api_key: Optional[Text] = None) -> Dict:
    @classmethod
    def create_asset_repo(
        cls,
        name: Text,
        description: Text,
        function: Text,
        source_language: Text,
        input_modality: Text,
        output_modality: Text,
        documentation_url: Optional[Text] = "",
        api_key: Optional[Text] = None,
    ) -> Dict:
        """Creates an image repository for this model and registers it in the
        platform backend.

        Args:
            name (Text): Model name
            hosting_machine (Text): Hosting machine ID obtained via list_host_machines
            always_on (bool): Whether the model should always be on
            version (Text): Model version
            description (Text): Model description
            function (Text): Model function name obtained via LIST_HOST_MACHINES
            is_async (bool): Whether the model is asynchronous or not (False in first release)
            source_language (Text): 2-character 639-1 code or 3-character 639-3 language code.
            api_key (Text, optional): Team API key. Defaults to None.

        Returns:
            Dict: Backend response
        """
        # Reconcile function name to be function ID in the backend
        function_list = cls.list_functions(True, config.TEAM_API_KEY)["items"]
        function_id = None
        for function_dict in function_list:
            if function_dict["name"] == function:
                function_id = function_dict["id"]
        if function_id is None:
            raise Exception(f"Invalid function name {function}")
        create_url = urljoin(config.BACKEND_URL, "sdk/models/onboard")
        logging.debug(f"URL: {create_url}")
        if api_key:
            headers = {"x-api-key": f"{api_key}", "Content-Type": "application/json"}
        else:
            headers = {"x-api-key": f"{config.TEAM_API_KEY}", "Content-Type": "application/json"}

        payload = {
            "model": {
                "name": name,
                "description": description,
                "connectionType": ["synchronous"],
                "function": function_id,
                "modalities": [f"{input_modality}-{output_modality}"],
                "documentationUrl": documentation_url,
                "sourceLanguage": source_language,
            },
            "source": "aixplain-ecr",
            "onboardingParams": {},
        }
        logging.debug(f"Body: {str(payload)}")
        response = _request_with_retry("post", create_url, headers=headers, json=payload)

        assert response.status_code == 201

        return response.json()

    @classmethod
    def asset_repo_login(cls, api_key: Optional[Text] = None) -> Dict:
        """Return login credentials for the image repository that corresponds with
        the given API_KEY.

        Args:
            api_key (Text, optional): Team API key. Defaults to None.

        Returns:
            Dict: Backend response
        """
        login_url = urljoin(config.BACKEND_URL, "sdk/ecr/login")
        logging.debug(f"URL: {login_url}")
        if api_key:
            headers = {"Authorization": f"Token {api_key}", "Content-Type": "application/json"}
        else:
            headers = {"Authorization": f"Token {config.TEAM_API_KEY}", "Content-Type": "application/json"}
        response = _request_with_retry("post", login_url, headers=headers)
        response_dict = json.loads(response.text)
        return response_dict

    @classmethod
    def onboard_model(
        cls,
        model_id: Text,
        image_tag: Text,
        image_hash: Text,
        host_machine: Optional[Text] = "",
        api_key: Optional[Text] = None,
    ) -> Dict:
        """Onboard a model after its image has been pushed to ECR.

        Args:
            model_id (Text): Model ID obtained from CREATE_ASSET_REPO.
            image_tag (Text): Image tag to be onboarded.
            image_hash (Text): Image digest.
            host_machine (Text, optional): Machine on which to host model.
            api_key (Text, optional): Team API key. Defaults to None.
        Returns:
            Dict: Backend response
        """
        onboard_url = urljoin(config.BACKEND_URL, f"sdk/models/{model_id}/onboarding")
        logging.debug(f"URL: {onboard_url}")
        if api_key:
            headers = {"x-api-key": f"{api_key}", "Content-Type": "application/json"}
        else:
            headers = {"x-api-key": f"{config.TEAM_API_KEY}", "Content-Type": "application/json"}
        payload = {"image": image_tag, "sha": image_hash, "hostMachine": host_machine}
        logging.debug(f"Body: {str(payload)}")
        response = _request_with_retry("post", onboard_url, headers=headers, json=payload)
        if response.status_code == 201:
            message = "Your onboarding request has been submitted to an aiXplain specialist for finalization. We will notify you when the process is completed."
            logging.info(message)
        else:
            message = "An error has occurred. Please make sure your model_id is valid and your host_machine, if set, is a valid option from the LIST_GPUS function."
        return response

    @classmethod
    def deploy_huggingface_model(
        cls,
        name: Text,
        hf_repo_id: Text,
        revision: Optional[Text] = "",
        hf_token: Optional[Text] = "",
        api_key: Optional[Text] = None,
    ) -> Dict:
        """Onboards and deploys a Hugging Face large language model.

        Args:
            name (Text): The user's name for the model.
            hf_repo_id (Text): The Hugging Face repository ID for this model ({author}/{model name}).
            hf_token (Text, optional): Hugging Face access token. Defaults to None.
            api_key (Text, optional): Team API key. Defaults to None.
        Returns:
            Dict: Backend response
        """
        supplier, model_name = hf_repo_id.split("/")
        deploy_url = urljoin(config.BACKEND_URL, "sdk/model-onboarding/onboard")
        if api_key:
            headers = {"Authorization": f"Token {api_key}", "Content-Type": "application/json"}
        else:
            headers = {"Authorization": f"Token {config.TEAM_API_KEY}", "Content-Type": "application/json"}
        body = {
            "model": {
                "name": name,
                "description": "A user-deployed Hugging Face model",
                "connectionType": ["synchronous"],
                "function": "text-generation",
                "documentationUrl": "aiXplain",
                "sourceLanguage": "en",
            },
            "source": "huggingface",
            "onboardingParams": {
                "hf_supplier": supplier,
                "hf_model_name": model_name,
                "hf_token": hf_token,
                "revision": revision,
            },
        }
        response = _request_with_retry("post", deploy_url, headers=headers, json=body)
        logging.debug(response.text)
        response_dicts = json.loads(response.text)
        return response_dicts

    @classmethod
    def get_huggingface_model_status(cls, model_id: Text, api_key: Optional[Text] = None):
        """Gets the on-boarding status of a Hugging Face model with ID MODEL_ID.

        Args:
            model_id (Text): The model's ID as returned by DEPLOY_HUGGINGFACE_MODEL
            api_key (Text, optional): Team API key. Defaults to None.
        Returns:
            Dict: Backend response
        """
        status_url = urljoin(config.BACKEND_URL, f"sdk/models/{model_id}")
        if api_key:
            headers = {"Authorization": f"Token {api_key}", "Content-Type": "application/json"}
        else:
            headers = {"Authorization": f"Token {config.TEAM_API_KEY}", "Content-Type": "application/json"}
        response = _request_with_retry("get", status_url, headers=headers)
        logging.debug(response.text)
        response_dicts = json.loads(response.text)
        ret_dict = {
            "status": response_dicts["status"],
            "name": response_dicts["name"],
            "id": response_dicts["id"],
            "pricing": response_dicts["pricing"],
        }
        return ret_dict



import pytest
import requests_mock
from aixplain.modules import Agent
from aixplain.utils import config
from aixplain.factories import AgentFactory
from aixplain.modules.agent import PipelineTool, ModelTool
from urllib.parse import urljoin


def test_fail_no_data_query():
    agent = Agent("123", "Test Agent")
    with pytest.raises(Exception) as exc_info:
        agent.run_async()
    assert str(exc_info.value) == "Either 'data' or 'query' must be provided."


def test_fail_query_must_be_provided():
    agent = Agent("123", "Test Agent")
    with pytest.raises(Exception) as exc_info:
        agent.run_async(data={})
    assert str(exc_info.value) == "When providing a dictionary, 'query' must be provided."


def test_fail_query_as_text_when_content_not_empty():
    agent = Agent("123", "Test Agent")
    with pytest.raises(Exception) as exc_info:
        agent.run_async(
            data={"query": "https://aixplain-platform-assets.s3.amazonaws.com/samples/en/CPAC1x2.wav"},
            content=["https://aixplain-platform-assets.s3.amazonaws.com/samples/en/CPAC1x2.wav"],
        )
    assert str(exc_info.value) == "When providing 'content', query must be text."


def test_fail_content_exceed_maximum():
    agent = Agent("123", "Test Agent")
    with pytest.raises(Exception) as exc_info:
        agent.run_async(
            data={"query": "Transcribe the audios:"},
            content=[
                "https://aixplain-platform-assets.s3.amazonaws.com/samples/en/CPAC1x2.wav",
                "https://aixplain-platform-assets.s3.amazonaws.com/samples/en/CPAC1x2.wav",
                "https://aixplain-platform-assets.s3.amazonaws.com/samples/en/CPAC1x2.wav",
                "https://aixplain-platform-assets.s3.amazonaws.com/samples/en/CPAC1x2.wav",
            ],
        )
    assert str(exc_info.value) == "The maximum number of content inputs is 3."


def test_fail_key_not_found():
    agent = Agent("123", "Test Agent")
    with pytest.raises(Exception) as exc_info:
        agent.run_async(data={"query": "Translate the text: {{input1}}"}, content={"input2": "Hello, how are you?"})
    assert str(exc_info.value) == "Key 'input2' not found in query."


def test_sucess_query_content():
    agent = Agent("123", "Test Agent")
    with requests_mock.Mocker() as mock:
        url = agent.url
        headers = {"x-api-key": config.TEAM_API_KEY, "Content-Type": "application/json"}
        ref_response = {"data": "Hello, how are you?", "status": "IN_PROGRESS"}
        mock.post(url, headers=headers, json=ref_response)

        response = agent.run_async(data={"query": "Translate the text: {{input1}}"}, content={"input1": "Hello, how are you?"})
    assert response["status"] == ref_response["status"]
    assert response["url"] == ref_response["data"]


def test_invalid_pipelinetool():
    with pytest.raises(Exception) as exc_info:
        AgentFactory.create(
            name="Test", tools=[PipelineTool(pipeline="309851793", description="Test")], llm_id="6646261c6eb563165658bbb1"
        )
    assert str(exc_info.value) == "Pipeline Tool Unavailable. Make sure Pipeline '309851793' exists or you have access to it."


def test_invalid_modeltool():
    with pytest.raises(Exception) as exc_info:
        AgentFactory.create(name="Test", tools=[ModelTool(model="309851793")], llm_id="6646261c6eb563165658bbb1")
    assert str(exc_info.value) == "Model Tool Unavailable. Make sure Model '309851793' exists or you have access to it."


def test_create_agent():
    from aixplain.enums import Supplier

    with requests_mock.Mocker() as mock:
        url = urljoin(config.BACKEND_URL, "sdk/agents")
        headers = {"x-api-key": config.TEAM_API_KEY, "Content-Type": "application/json"}

        ref_response = {
            "id": "123",
            "name": "Test Agent",
            "description": "Test Agent Description",
            "teamId": "123",
            "version": "1.0",
            "status": "onboarded",
            "llmId": "6646261c6eb563165658bbb1",
            "pricing": {"currency": "USD", "value": 0.0},
            "assets": [
                {
                    "type": "model",
                    "supplier": "openai",
                    "version": "1.0",
                    "assetId": "6646261c6eb563165658bbb1",
                    "function": "text-generation",
                }
            ],
        }
        mock.post(url, headers=headers, json=ref_response)

        url = urljoin(config.BACKEND_URL, "sdk/models/6646261c6eb563165658bbb1")
        model_ref_response = {
            "id": "6646261c6eb563165658bbb1",
            "name": "Test LLM",
            "description": "Test LLM Description",
            "function": {"id": "text-generation"},
            "supplier": "openai",
            "version": {"id": "1.0"},
            "status": "onboarded",
            "pricing": {"currency": "USD", "value": 0.0},
        }
        mock.get(url, headers=headers, json=model_ref_response)

        agent = AgentFactory.create(
            name="Test Agent",
            description="Test Agent Description",
            llm_id="6646261c6eb563165658bbb1",
            tools=[AgentFactory.create_model_tool(supplier=Supplier.OPENAI, function="text-generation")],
        )

    assert agent.name == ref_response["name"]
    assert agent.description == ref_response["description"]
    assert agent.llm_id == ref_response["llmId"]
    assert agent.tools[0].function.value == ref_response["assets"][0]["function"]



from pathlib import Path
from typing import Optional, Text, Union, Dict, List, Any
from enum import Enum

from aixplain.modules.metadata import MetaData
from aixplain.enums.function import FunctionInputOutput, Function
from aixplain.enums.data_type import DataType
from aixplain.enums import DataSubtype


def _is_split(dsubtype: Union[Text, DataSubtype]):
    return str(dsubtype).lower() == "split"


def dataset_onboarding_validation(
    input_schema: List[Union[Dict, MetaData]],
    output_schema: List[Union[Dict, MetaData]],
    function: Function,
    input_ref_data: Dict[Text, Any] = {},
    metadata_schema: List[Union[Dict, MetaData]] = [],
    content_path: Union[Union[Text, Path], List[Union[Text, Path]]] = [],
    split_labels: Optional[List[Text]] = None,
    split_rate: Optional[List[float]] = None,
    s3_link: Optional[str] = None,
) -> None:
    """Dataset Onboard Validation
    Args:
        input_schema (List[Union[Dict, MetaData]]): metadata of inputs
        output_schema (List[Union[Dict, MetaData]]): metadata of outputs
        function (Function): dataset function
        input_ref_data (Dict[Text, Any], optional): reference to input data which is already in the platform. Defaults to {}.
        metadata_schema (List[Union[Dict, MetaData]], optional): metadata of metadata information of the dataset. Defaults to [].
        content_path (Union[Union[Text, Path], List[Union[Text, Path]]]): path to files which contain the data content
        split_labels: (Optional[List[Text]]): The delimiters according which to split the dataset
        split_rate: (Optional[List[float]]): the rate of spliting the dataset
        s3_link (Optional[str]): s3 url to files or directories
    """

    metadata_spliting_schema = list(filter(lambda md: str(md.dsubtype) == "split", metadata_schema))

    # validate the input and the output of the dataset
    assert (
        len(input_schema) > 0 or len(input_ref_data) > 0
    ), "Data Asset Onboarding Error: You must specify an input data to onboard a dataset."

    input_dtype = input_schema[0].dtype if isinstance(input_schema[0], MetaData) else input_schema[0]["dtype"]
    if isinstance(input_dtype, DataType):
        input_dtype = input_dtype.value

    assert (
        FunctionInputOutput.get(function) is not None and input_dtype in FunctionInputOutput[function]["input"]
    ), f"Data Asset Onboarding Error: The input data type `{input_dtype}` is not compatible with the `{function}` function.\nThe expected input data type should be one of these data type: `{FunctionInputOutput[function]['input']}`."

    if len(output_schema) > 0:
        output_dtype = output_schema[0].dtype if isinstance(output_schema[0], MetaData) else output_schema[0]["dtype"]
        if isinstance(output_dtype, DataType):
            output_dtype = output_dtype.value

        assert (
            FunctionInputOutput.get(function) is not None and output_dtype in FunctionInputOutput[function]["output"]
        ), f"Data Asset Onboarding Error: The output data type `{output_dtype}` is not compatible with the `{function}` function.\nThe expected output data type should be one of these data type: `{FunctionInputOutput[function]['output']}`."

    # validate the splitting
    assert (
        len(metadata_spliting_schema) < 2
    ), f"Data Asset Onboarding Error: Only 0 or 1 metadata of the split subtype can be added to the `metadata_schema`."
    assert all(
        str(mds.dtype) == "label" for mds in metadata_spliting_schema
    ), f"Data Asset Onboarding Error: The `dtype` must be `label` for any splitting subtype."

    assert (
        content_path is not None or s3_link is not None
    ), "Data Asset Onboarding Error: No path to content Data was provided. Please update `context_path` or `s3_link`."
    assert (split_labels is not None and split_rate is not None) or (
        split_labels is None and split_rate is None
    ), "Data Asset Onboarding Error: Make sure you set the split labels values as well as their rates."



__author__ = "aiXplain"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Michael Lam
Date: September 18th 2023
Description:
    Model Factory CLI
"""

from aixplain.factories.model_factory import ModelFactory
from typing import Dict, List, Optional, Text
import click
import yaml

@click.command("hosts")
@click.option("--api-key", default=None, 
              help="TEAM_API_KEY if not already set in environment")
def list_host_machines(api_key: Optional[Text] = None) -> None:
    """CLI wrapper function for the LIST_HOST_MACHINES function in 
    ModelFactory.

    Args:
        api_key (Text, optional): Team API key. Defaults to None.

    Returns:
        None
    """
    ret_val = ModelFactory.list_host_machines(api_key)
    ret_val_yaml = yaml.dump(ret_val)
    click.echo(ret_val_yaml)

@click.command("functions")
@click.option("--verbose", is_flag=True, 
              help="List all function details, False by default.")
@click.option("--api-key", default=None, 
              help="TEAM_API_KEY if not already set in environment.")
def list_functions(verbose: bool, api_key: Optional[Text] = None) -> None:
    """CLI wrapper function for the LIST_FUNCTIONS function in ModelFactory.

    Args:
        verbose (Boolean, optional): Set to True if a detailed response 
            is desired; is otherwise False by default.
        api_key (Text, optional): Team API key. Defaults to None.
    Returns:
        None
    """
    ret_val = ModelFactory.list_functions(verbose, api_key)
    ret_val_yaml = yaml.dump(ret_val)
    click.echo(ret_val_yaml)

@click.command("gpus")
@click.option("--api-key", default=None, 
              help="TEAM_API_KEY if not already set in environment.")
def list_gpus(api_key: Optional[Text] = None) -> None:
    """CLI wrapper function for the LIST_GPUS function in ModelFactory.

    Args:
        api_key (Text, optional): Team API key. Defaults to None.
    Returns:
        None
    """
    ret_val = ModelFactory.list_gpus(api_key)
    ret_val_yaml = yaml.dump(ret_val)
    click.echo(ret_val_yaml)

@click.command("image-repo")
@click.option("--name", help="Model name.")
@click.option("--description", help="Description of model.")
@click.option("--function", help="Function name obtained from LIST_FUNCTIONS.")
@click.option("--source-language", default="en", 
              help="Model source language in 2-character 639-1 code or 3-character 639-3 code.")
@click.option("--input-modality", help="Input type (text, video, image, etc.)")
@click.option("--output-modality", help="Output type (text, video, image, etc.)")
@click.option("--documentation-url", default="", help="Link to model documentation.")
@click.option("--api-key", default=None, 
              help="TEAM_API_KEY if not already set in environment.")
def create_asset_repo(name: Text, description: Text, function: Text,
                      source_language: Text, input_modality: Text, 
                      output_modality: Text, 
                      documentation_url: Optional[Text] = "",
                      api_key: Optional[Text] = None) -> None:
    """CLI wrapper function for the CREATE_ASSET_REPO function in ModelFactory.

    Args:
        name (Text): Model name
        hosting_machine (Text): Hosting machine ID obtained via list_host_machines
        always_on (bool): Whether the model should always be on
        version (Text): Model version
        description (Text): Model description
        function (Text): Model function name obtained via LIST_HOST_MACHINES
        is_async (bool): Whether the model is asynchronous or not (False in first release)
        source_language (Text): 2-character 639-1 code or 3-character 639-3 language code.
        api_key (Text, optional): Team API key. Defaults to None.

    Returns:
        None
    """
    ret_val = ModelFactory.create_asset_repo(name, description, function, 
                                             source_language, input_modality, 
                                             output_modality, documentation_url, 
                                             api_key)
    ret_val_yaml = yaml.dump(ret_val)
    click.echo(ret_val_yaml)

@click.command("image-repo-login")
@click.option("--api-key", default=None, 
              help="TEAM_API_KEY if not already set in environment.")
def asset_repo_login(api_key: Optional[Text] = None) -> None:
    """CLI wrapper function for the ASSET_REPO_LOGIN function in ModelFactory.

    Args:
        api_key (Text, optional): Team API key. Defaults to None.

    Returns:
        None
    """
    ret_val = ModelFactory.asset_repo_login(api_key)
    ret_val_yaml = yaml.dump(ret_val)
    click.echo(ret_val_yaml)

@click.command("model")
@click.option("--model-id", help="Model ID from CREATE_IMAGE_REPO.")
@click.option("--image-tag", help="The tag of the image that you would like hosted.")
@click.option("--image-hash", help="The hash of the image you would like onboarded.")
@click.option("--host-machine", default="", help="The machine on which to host the model.")
@click.option("--api-key", default=None, help="TEAM_API_KEY if not already set in environment.")
def onboard_model(model_id: Text, image_tag: Text, image_hash: Text, 
                  host_machine: Optional[Text] = "", 
                  api_key: Optional[Text] = None) -> None:
    """CLI wrapper function for the ONBOARD_MODEL function in ModelFactory.

    Args:
        model_id (Text): Model ID obtained from CREATE_ASSET_REPO.
        image_tag (Text): Image tag to be onboarded.
        api_key (Text, optional): Team API key. Defaults to None.

    Returns:
        None
    """   
    ret_val = ModelFactory.onboard_model(model_id, image_tag, image_hash, 
                                         host_machine, api_key)
    ret_val_yaml = yaml.dump(ret_val)
    click.echo(ret_val_yaml)

@click.command("hf-model")
@click.option("--name", help="User-defined name for Hugging Face model.")
@click.option("--hf-repo-id", help="Repository ID from Hugging Face in {supplier}/{model name} form.")
@click.option("--revision", default="", help="Commit hash of repository.")
@click.option("--hf-token", default=None, help="Hugging Face token used to authenticate to this model.")
@click.option("--api-key", default=None, help="TEAM_API_KEY if not already set in environment.")
def deploy_huggingface_model(name: Text, hf_repo_id: Text, 
                             hf_token: Optional[Text] = None, 
                             revision: Optional[Text] = None,
                             api_key: Optional[Text] = None) -> None:
    """CLI wrapper function for the DEPLOY_HUGGINGFACE_MODEL function in ModelFactory.

    Args:
        name (Text): User-defined name for Hugging Face model.
        api_key (Text, optional): Team API key. Defaults to None.

    Returns:
        None
    """
    ret_val = ModelFactory.deploy_huggingface_model(name, hf_repo_id, revision, hf_token, api_key)
    ret_val_yaml = yaml.dump(ret_val)
    click.echo(ret_val_yaml)

@click.command("hf-model-status")
@click.option("--model-id", help="Model ID from DEPLOY_HUGGINGFACE_MODEL.")
@click.option("--api-key", default=None, help="TEAM_API_KEY if not already set in environment.")
def get_huggingface_model_status(model_id: Text, api_key: Optional[Text] = None) -> None:
    """CLI wrapper function for the GET_HUGGINGFACE_MODEL_STATUS function in ModelFactory.

    Args:
        model_id (Text): Model ID obtained from DEPLOY_HUGGINGFACE_MODEL.
        api_key (Text, optional): Team API key. Defaults to None.

    Returns:
        None
    """
    ret_val = ModelFactory.get_huggingface_model_status(model_id, api_key)
    ret_val_yaml = yaml.dump(ret_val)
    click.echo(ret_val_yaml)



"""
aiXplain SDK Library.
---

aiXplain SDK enables python programmers to add AI functions
to their software.

Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from .asset_factory import AssetFactory
from .agent_factory import AgentFactory
from .benchmark_factory import BenchmarkFactory
from .corpus_factory import CorpusFactory
from .data_factory import DataFactory
from .dataset_factory import DatasetFactory
from .file_factory import FileFactory
from .metric_factory import MetricFactory
from .model_factory import ModelFactory
from .pipeline_factory import PipelineFactory
from .finetune_factory import FinetuneFactory
from .wallet_factory import WalletFactory



__author__ = "thiagocastroferreira"

"""
Copyright 2023 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: aiXplain team
Date: June 6th 2023
Description:
    Content Interval
"""

from dataclasses import dataclass
from typing import List, Optional, Text, Tuple, Union


@dataclass
class ContentInterval:
    content: Text
    content_id: int


@dataclass
class TextContentInterval(ContentInterval):
    start: Union[int, Tuple[int, int]]
    end: Union[int, Tuple[int, int]]


@dataclass
class AudioContentInterval(ContentInterval):
    start_time: float
    end_time: float


@dataclass
class ImageContentInterval(ContentInterval):
    x: Union[float, List[float]]
    y: Union[float, List[float]]
    width: Optional[float] = None
    height: Optional[float] = None
    rotation: Optional[float] = None


@dataclass
class VideoContentInterval(ContentInterval):
    start_time: float
    end_time: float
    x: Optional[Union[float, List[float]]] = None
    y: Optional[Union[float, List[float]]] = None
    width: Optional[float] = None
    height: Optional[float] = None
    rotation: Optional[float] = None



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from enum import Enum
from typing import Text


class SchedulerType(Text, Enum):
    LINEAR = "linear"
    COSINE = "cosine"
    COSINE_WITH_RESTARTS = "cosine_with_restarts"
    POLYNOMIAL = "polynomial"
    CONSTANT = "constant"
    CONSTANT_WITH_WARMUP = "constant_with_warmup"
    INVERSE_SQRT = "inverse_sqrt"
    REDUCE_ON_PLATEAU = "reduce_lr_on_plateau"


EPOCHS_MAX_VALUE = 4
BATCH_SIZE_VALUES = [1, 2, 4, 8, 16, 32, 64]
MAX_SEQ_LENGTH_MAX_VALUE = 4096


@dataclass_json
@dataclass
class Hyperparameters(object):
    epochs: int = 1
    train_batch_size: int = 4
    eval_batch_size: int = 4
    learning_rate: float = 1e-5
    max_seq_length: int = 4096
    warmup_ratio: float = 0.0
    warmup_steps: int = 0
    lr_scheduler_type: SchedulerType = SchedulerType.LINEAR

    def __post_init__(self):
        if not isinstance(self.epochs, int):
            raise TypeError("epochs should be of type int")

        if not isinstance(self.train_batch_size, int):
            raise TypeError("train_batch_size should be of type int")

        if not isinstance(self.eval_batch_size, int):
            raise TypeError("eval_batch_size should be of type int")

        if not isinstance(self.learning_rate, float):
            raise TypeError("learning_rate should be of type float")

        if not isinstance(self.max_seq_length, int):
            raise TypeError("max_seq_length should be of type int")

        if not isinstance(self.warmup_ratio, float):
            raise TypeError("warmup_ratio should be of type float")

        if not isinstance(self.warmup_steps, int):
            raise TypeError("warmup_steps should be of type int")

        if not isinstance(self.lr_scheduler_type, SchedulerType):
            raise TypeError("lr_scheduler_type should be of type SchedulerType")

        if self.epochs > EPOCHS_MAX_VALUE:
            raise ValueError(f"epochs must be less or equal to {EPOCHS_MAX_VALUE}")

        if self.train_batch_size not in BATCH_SIZE_VALUES:
            raise ValueError(f"train_batch_size must be one of the following values: {BATCH_SIZE_VALUES}")

        if self.eval_batch_size not in BATCH_SIZE_VALUES:
            raise ValueError(f"eval_batch_size must be one of the following values: {BATCH_SIZE_VALUES}")

        if self.max_seq_length > MAX_SEQ_LENGTH_MAX_VALUE:
            raise ValueError(f"max_seq_length must be less or equal to {MAX_SEQ_LENGTH_MAX_VALUE}")



from typing import List, Text
from aixplain.modules.dataset import Dataset
import re


def _get_data_list(dataset: Dataset):
    flatten_target_values = [item for sublist in list(dataset.target_data.values()) for item in sublist]
    data_list = list(dataset.source_data.values()) + flatten_target_values
    return data_list


def validate_prompt(prompt: Text, dataset_list: List[Dataset]) -> Text:
    result_prompt = prompt
    referenced_data = set(re.findall("<<(.+?)>>", prompt))
    for dataset in dataset_list:
        data_list = _get_data_list(dataset)
        for data in data_list:
            if data.id in referenced_data:
                result_prompt = result_prompt.replace(f"<<{data.id}>>", f"<<{data.name}>>")
                referenced_data.remove(data.id)
                referenced_data.add(data.name)

    # check if dataset list has same data name and it is referenced
    name_set = set()
    for dataset in dataset_list:
        data_list = _get_data_list(dataset)
        for data in data_list:
            assert not (
                data.name in name_set and data.name in referenced_data
            ), "Datasets must not have more than one referenced data with same name"
            name_set.add(data.name)

    # check if all referenced data have a respective data in dataset list
    for dataset in dataset_list:
        data_list = _get_data_list(dataset)
        for data in data_list:
            if data.name in referenced_data:
                result_prompt = result_prompt.replace(f"<<{data.name}>>", f"{{{data.name}}}")
                referenced_data.remove(data.name)
    assert len(referenced_data) == 0, "Referenced data are not present in dataset list"
    return result_prompt



__author__ = "aiXplain"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: aiXplain team
Date: March 20th 2023
Description:
    Data Class
"""

from aixplain.modules.file import File
from aixplain.enums.data_subtype import DataSubtype
from aixplain.enums.data_type import DataType
from aixplain.enums.language import Language
from aixplain.enums.onboard_status import OnboardStatus
from aixplain.enums.privacy import Privacy
from typing import List, Optional, Text, Any


class Data:
    def __init__(
        self,
        id: Text,
        name: Text,
        dtype: DataType,
        privacy: Privacy,
        onboard_status: OnboardStatus,
        data_column: Optional[Any] = None,
        start_column: Optional[Any] = None,
        end_column: Optional[Any] = None,
        files: List[File] = [],
        languages: List[Language] = [],
        dsubtype: DataSubtype = DataSubtype.OTHER,
        length: Optional[int] = None,
        **kwargs
    ) -> None:
        """Data Class.

        Description:
            Data consists of a list of samples of same type and genre.

        Args:
            id (Text): Data ID
            name (Text): Data Name
            dtype (DataType): Data Type
            privacy (Privacy): Privacy of data
            onboard_status (OnboardStatus): onboard status
            data_column (Optional[Any], optional): Column index/name where the data is on a structured file (e.g. CSV). Defaults to None.
            start_column (Optional[Any], optional): Column index/name where the start indexes is on a structured file (e.g. CSV). Defaults to None.
            end_column (Optional[Any], optional): Column index/name where the end indexes is on a structured file (e.g. CSV). Defaults to None.
            files (List[File], optional): List of files where the data instances are stored. Defaults to [].
            languages (List[Language], optional): List of languages which the data consists of. Defaults to [].
            dsubtype (DataSubtype, optional): Data subtype (e.g., age, topic, race, split, etc.), used in datasets metadata. Defaults to Other.
            length (Optional[int], optional): Number of rows in the Data. Defaults to None.
        """
        self.id = id
        self.name = name
        self.dtype = dtype
        self.privacy = privacy
        if isinstance(onboard_status, str):
            onboard_status = OnboardStatus(onboard_status)
        self.onboard_status = onboard_status
        self.files = files
        if data_column is None:
            self.data_column = name
        else:
            self.data_column = data_column
        self.start_column = start_column
        self.end_column = end_column
        self.languages = []
        for language in languages:
            if isinstance(language, str):
                language = Language(language)
            self.languages.append(language)
        self.dsubtype = dsubtype
        self.length = length
        self.kwargs = kwargs



__author__ = "aiXplain"

"""
Copyright 2024 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Lucas Pavanelli and Thiago Castro Ferreira
Date: May 16th 2024
Description:
    Agentification Class
"""
import json
import logging
import time
import traceback

from aixplain.utils.file_utils import _request_with_retry
from aixplain.enums.supplier import Supplier
from aixplain.enums.asset_status import AssetStatus
from aixplain.enums.storage_type import StorageType
from aixplain.modules.model import Model
from aixplain.modules.agent.tool import Tool
from aixplain.modules.agent.tool.model_tool import ModelTool
from aixplain.modules.agent.tool.pipeline_tool import PipelineTool
from typing import Dict, List, Text, Optional, Union
from urllib.parse import urljoin

from aixplain.utils import config


class Agent(Model):
    """Advanced AI system capable of performing tasks by leveraging specialized software tools and resources from aiXplain marketplace.

    Attributes:
        id (Text): ID of the Agent
        name (Text): Name of the Agent
        tools (List[Tool]): List of tools that the Agent uses.
        description (Text, optional): description of the Agent. Defaults to "".
        llm_id (Text, optional): large language model. Defaults to GPT-4o (6646261c6eb563165658bbb1).
        supplier (Text): Supplier of the Agent.
        version (Text): Version of the Agent.
        backend_url (str): URL of the backend.
        api_key (str): The TEAM API key used for authentication.
        cost (Dict, optional): model price. Defaults to None.
    """

    def __init__(
        self,
        id: Text,
        name: Text,
        tools: List[Tool] = [],
        description: Text = "",
        llm_id: Text = "6646261c6eb563165658bbb1",
        api_key: Optional[Text] = config.TEAM_API_KEY,
        supplier: Union[Dict, Text, Supplier, int] = "aiXplain",
        version: Optional[Text] = None,
        cost: Optional[Dict] = None,
        status: AssetStatus = AssetStatus.ONBOARDING,
        **additional_info,
    ) -> None:
        """Create a FineTune with the necessary information.

        Args:
            id (Text): ID of the Agent
            name (Text): Name of the Agent
            tools (List[Tool]): List of tools that the Agent uses.
            description (Text, optional): description of the Agent. Defaults to "".
            llm_id (Text, optional): large language model. Defaults to GPT-4o (6646261c6eb563165658bbb1).
            supplier (Text): Supplier of the Agent.
            version (Text): Version of the Agent.
            backend_url (str): URL of the backend.
            api_key (str): The TEAM API key used for authentication.
            cost (Dict, optional): model price. Defaults to None.
        """
        # assert len(tools) > 0, "At least one tool must be provided."
        super().__init__(id, name, description, api_key, supplier, version, cost=cost)
        self.additional_info = additional_info
        self.tools = tools
        self.llm_id = llm_id
        if isinstance(status, str):
            try:
                status = AssetStatus(status)
            except Exception:
                status = AssetStatus.ONBOARDING
        self.status = status

    def run(
        self,
        data: Optional[Union[Dict, Text]] = None,
        query: Optional[Text] = None,
        session_id: Optional[Text] = None,
        history: Optional[List[Dict]] = None,
        name: Text = "model_process",
        timeout: float = 300,
        parameters: Dict = {},
        wait_time: float = 0.5,
        content: Optional[Union[Dict[Text, Text], List[Text]]] = None,
    ) -> Dict:
        """Runs an agent call.

        Args:
            data (Optional[Union[Dict, Text]], optional): data to be processed by the agent. Defaults to None.
            query (Optional[Text], optional): query to be processed by the agent. Defaults to None.
            session_id (Optional[Text], optional): conversation Session ID. Defaults to None.
            history (Optional[List[Dict]], optional): chat history (in case session ID is None). Defaults to None.
            name (Text, optional): ID given to a call. Defaults to "model_process".
            timeout (float, optional): total polling time. Defaults to 300.
            parameters (Dict, optional): optional parameters to the model. Defaults to "{}".
            wait_time (float, optional): wait time in seconds between polling calls. Defaults to 0.5.
            content (Union[Dict[Text, Text], List[Text]], optional): Content inputs to be processed according to the query. Defaults to None.

        Returns:
            Dict: parsed output from model
        """
        start = time.time()
        try:
            response = self.run_async(
                data=data,
                query=query,
                session_id=session_id,
                history=history,
                name=name,
                parameters=parameters,
                content=content,
            )
            if response["status"] == "FAILED":
                end = time.time()
                response["elapsed_time"] = end - start
                return response
            poll_url = response["url"]
            end = time.time()
            response = self.sync_poll(poll_url, name=name, timeout=timeout, wait_time=wait_time)
            return response
        except Exception as e:
            msg = f"Error in request for {name} - {traceback.format_exc()}"
            logging.error(f"Model Run: Error in running for {name}: {e}")
            end = time.time()
            return {"status": "FAILED", "error": msg, "elapsed_time": end - start}

    def run_async(
        self,
        data: Optional[Union[Dict, Text]] = None,
        query: Optional[Text] = None,
        session_id: Optional[Text] = None,
        history: Optional[List[Dict]] = None,
        name: Text = "model_process",
        parameters: Dict = {},
        content: Optional[Union[Dict[Text, Text], List[Text]]] = None,
    ) -> Dict:
        """Runs asynchronously an agent call.

        Args:
            data (Optional[Union[Dict, Text]], optional): data to be processed by the agent. Defaults to None.
            query (Optional[Text], optional): query to be processed by the agent. Defaults to None.
            session_id (Optional[Text], optional): conversation Session ID. Defaults to None.
            history (Optional[List[Dict]], optional): chat history (in case session ID is None). Defaults to None.
            name (Text, optional): ID given to a call. Defaults to "model_process".
            parameters (Dict, optional): optional parameters to the model. Defaults to "{}".
            content (Union[Dict[Text, Text], List[Text]], optional): Content inputs to be processed according to the query. Defaults to None.

        Returns:
            dict: polling URL in response
        """
        from aixplain.factories.file_factory import FileFactory

        assert data is not None or query is not None, "Either 'data' or 'query' must be provided."
        if data is not None:
            if isinstance(data, dict):
                assert "query" in data and data["query"] is not None, "When providing a dictionary, 'query' must be provided."
                query = data.get("query")
                if session_id is None:
                    session_id = data.get("session_id")
                if history is None:
                    history = data.get("history")
                if content is None:
                    content = data.get("content")
            else:
                query = data

        # process content inputs
        if content is not None:
            assert FileFactory.check_storage_type(query) == StorageType.TEXT, "When providing 'content', query must be text."

            if isinstance(content, list):
                assert len(content) <= 3, "The maximum number of content inputs is 3."
                for input_link in content:
                    input_link = FileFactory.to_link(input_link)
                    query += f"\n{input_link}"
            elif isinstance(content, dict):
                for key, value in content.items():
                    assert "{{" + key + "}}" in query, f"Key '{key}' not found in query."
                    value = FileFactory.to_link(value)
                    query = query.replace("{{" + key + "}}", f"'{value}'")

        headers = {"x-api-key": self.api_key, "Content-Type": "application/json"}

        payload = {"id": self.id, "query": FileFactory.to_link(query), "sessionId": session_id, "history": history}
        payload.update(parameters)
        payload = json.dumps(payload)

        r = _request_with_retry("post", self.url, headers=headers, data=payload)
        logging.info(f"Model Run Async: Start service for {name} - {self.url} - {payload} - {headers}")

        resp = None
        try:
            resp = r.json()
            logging.info(f"Result of request for {name} - {r.status_code} - {resp}")

            poll_url = resp["data"]
            response = {"status": "IN_PROGRESS", "url": poll_url}
        except Exception:
            response = {"status": "FAILED"}
            msg = f"Error in request for {name} - {traceback.format_exc()}"
            logging.error(f"Model Run Async: Error in running for {name}: {resp}")
            if resp is not None:
                response["error"] = msg
        return response

    def delete(self) -> None:
        """Delete Corpus service"""
        try:
            url = urljoin(config.BACKEND_URL, f"sdk/agents/{self.id}")
            headers = {"x-api-key": config.TEAM_API_KEY, "Content-Type": "application/json"}
            logging.debug(f"Start service for DELETE Agent  - {url} - {headers}")
            r = _request_with_retry("delete", url, headers=headers)
            if r.status_code != 200:
                raise Exception()
        except Exception:
            message = f"Agent Deletion Error (HTTP {r.status_code}): Make sure the agent exists and you are the owner."
            logging.error(message)
            raise Exception(f"{message}")



__author__ = "thiagocastroferreira"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Duraikrishna Selvaraju and Thiago Castro Ferreira
Date: May 10th 2022
Description:
    Script to generate English Subtitles to Portuguese Videos

Notes:
    Example:
        python3 subtitle_generator.py --video-pt-path https://aixplain-samples-public.s3.amazonaws.com/pipelines/sample_subtitle.mp4 \
                                      --srt-path pt.srt
                                      -k <API_KEY>
"""

import argparse
from datetime import timedelta
import os
import logging

import requests
from tqdm import tqdm

from aixplain.factories.pipeline_factory import PipelineFactory


def generate_srt(segments: list, write_file: str = None):
    """
    Generate a subtitle .srt file

    params:
    ---
        segments: list of segments returned by pipeline execution. Each segment consists of:
            start_sec: start of the segment (in seconds)
            end_sec: end of the segment (in seconds)
            response: link to the transcription of the audio in the interval
        write_file: file to write the srt file. If not provided (None), the string will be returned
    return:
    ---
        .srt string or file
    """
    subtitles = []
    for i in tqdm(range(len(segments))):
        s = segments[i]
        start = str(timedelta(seconds=s["start_sec"]))
        if len(start.split(":")[0]) == 1:
            start = "0" + start
        start = start.replace(".", ",")[:12]

        end = str(timedelta(seconds=s["end_sec"]))
        if len(end.split(":")[0]) == 1:
            end = "0" + end
        end = end.replace(".", ",")[:12]

        r = requests.get(s["response"])
        text = r.text

        subtitles.append({"index": str(i + 1), "start": start, "end": end, "text": text})

    srt = []
    for s in subtitles:
        chunk = f"{s['index']}\n{s['start']} --> {s['end']}\n{s['text']}"
        srt.append(chunk)

    result = "\n\n".join(srt)
    if write_file is not None:
        with open(write_file, "w") as f:
            f.write(result)
    return result


def main(video_path: str, srt_path: str, pipeline_id: str):
    """
    params:
    ---
        video_path:str - path to the Portuguese video to be subtitled
        srt_path:str - path to save the generated subtitle
        pipeline_id:str - aixplain Pipeline ID for subtitling pipeline
    return:
    ---
        .srt string or file
    """
    pipe = PipelineFactory.get(pipeline_id)
    response = pipe.run(data=video_path)

    if response["success"] is True:
        if response["response"]["status"] == "SUCCESS":
            translation_outcome = [data for data in response["response"]["data"] if data["function"] == "translation"][0]
            segments = translation_outcome["segments"]

            generate_srt(segments, srt_path)
        else:
            raise Exception("Failing in getting the transcriptions to the segments of the Portuguese Video.")
    else:
        raise Exception("Failing in getting the transcriptions to the segments of the Portuguese Video.")


if __name__ == "__main__":
    # Create the parser
    parser = argparse.ArgumentParser()
    # Add arguments
    parser.add_argument("--video-pt-path", type=str, required=True)
    parser.add_argument("--srt-path", type=str, required=True)
    parser.add_argument("-k", "--pipeline-id", type=str, required=True)

    # Parse the argument
    args = parser.parse_args()

    video_path = args.video_pt_path
    srt_path = args.srt_path
    pipeline_id = args.pipeline_id

    logging.basicConfig(level=os.environ.get("LOG_LEVEL", "INFO"))

    main(video_path, srt_path, pipeline_id)



import os
import json
from typing import Tuple

import requests

from aixplain.utils import config


class ScriptFactory:

    @classmethod
    def upload_script(cls, script_path: str) -> Tuple[str, str]:
        try:
            url = f"{config.BACKEND_URL}/sdk/pipelines/script"
            headers = {"Authorization": f"Token {config.TEAM_API_KEY}"}
            r = requests.post(
                url, headers=headers, files={"file": open(script_path, "rb")}
            )
            if 200 <= r.status_code < 300:
                response = r.json()
            else:
                raise Exception()
        except Exception:
            response = {"fileId": ""}

        # get metadata info
        fname = os.path.splitext(os.path.basename(script_path))[0]
        file_size_kb = int(os.path.getsize(script_path) / 1024)
        metadata = json.dumps({"name": fname, "size": file_size_kb})
        return response["fileId"], metadata



__author__ = "aixplain"
import logging

from aixplain.enums import DataType
import aixplain.utils.config as config
from aixplain.modules.pipeline import Pipeline
from aixplain.modules.pipeline.designer import (
    Input,
    Output,
    AssetNode,
    Decision,
    Router,
    Route,
    Script,
    Link,
)
from typing import Dict


def build_from_response(response: Dict, load_architecture: bool = False) -> Pipeline:
    """Converts response Json to 'Pipeline' object

    Args:
        response (Dict): Json from API
        load_architecture (bool, optional): If True, the architecture will be loaded. Defaults to False.

    Returns:
        Pipeline: Coverted 'Pipeline' object
    """
    if "api_key" not in response:
        response["api_key"] = config.TEAM_API_KEY

    # instantiating pipeline generic info
    pipeline = Pipeline(response["id"], response["name"], response["api_key"])
    if load_architecture is True:
        try:
            # instantiating nodes
            for node_json in response["nodes"]:
                print(node_json)
                if node_json["type"].lower() == "input":
                    node = Input(
                        data=node_json["data"] if "data" in node_json else None,
                        data_types=[DataType(dt) for dt in node_json["dataType"]],
                    )
                elif node_json["type"].lower() == "asset":
                    node = AssetNode(asset_id=node_json["assetId"])
                elif node_json["type"].lower() == "segmentor":
                    raise NotImplementedError()
                elif node_json["type"].lower() == "reconstructor":
                    raise NotImplementedError()
                elif node_json["type"].lower() == "decision":
                    node = Decision(routes=[Route(**route) for route in node_json["routes"]])
                elif node_json["type"].lower() == "router":
                    node = Router(routes=[Route(**route) for route in node_json["routes"]])
                elif node_json["type"].lower() == "script":
                    node = Script(fileId=node_json["fileId"])
                elif node_json["type"].lower() == "output":
                    node = Output()

                if "inputValues" in node_json:
                    [
                        node.inputs.create_param(
                            data_type=DataType(input_param["dataType"]) if "dataType" in input_param else None,
                            code=input_param["code"],
                            value=input_param["value"] if "value" in input_param else None,
                            is_required=input_param["isRequired"] if "isRequired" in input_param else False,
                        )
                        for input_param in node_json["inputValues"]
                        if input_param["code"] not in node.inputs
                    ]
                if "outputValues" in node_json:
                    [
                        node.outputs.create_param(
                            data_type=DataType(output_param["dataType"]) if "dataType" in output_param else None,
                            code=output_param["code"],
                            value=output_param["value"] if "value" in output_param else None,
                            is_required=output_param["isRequired"] if "isRequired" in output_param else False,
                        )
                        for output_param in node_json["outputValues"]
                        if output_param["code"] not in node.outputs
                    ]
                node.number = node_json["number"]
                node.label = node_json["label"]
                pipeline.add_node(node)

            # instantiating links
            for link_json in response["links"]:
                for param_mapping in link_json["paramMapping"]:
                    link = Link(
                        from_node=pipeline.get_node(link_json["from"]),
                        to_node=pipeline.get_node(link_json["to"]),
                        from_param=param_mapping["from"],
                        to_param=param_mapping["to"],
                    )
                    pipeline.add_link(link)
        except Exception as e:
            logging.warning("Error loading pipeline architecture:, error: %s", e)
            pipeline.nodes = []
            pipeline.links = []
    return pipeline



import pathlib

import requests
from urllib.parse import urljoin
from jinja2 import Environment, BaseLoader

from aixplain.utils import config

SEGMENTOR_FUNCTIONS = [
    "split-on-linebreak",
    "speaker-diarization-audio",
    "voice-activity-detection",
]

RECONSTRUCTOR_FUNCTIONS = ["text-reconstruction", "audio-reconstruction"]

MODULE_NAME = "pipeline"
TEMPLATE = """# This is an auto generated module. PLEASE DO NOT EDIT


from typing import Union, Type
from aixplain.enums import DataType

from .designer import (
    InputParam,
    OutputParam,
    Inputs,
    Outputs,
    TI,
    TO,
    AssetNode,
    BaseReconstructor,
    BaseSegmentor,
)
from .default import DefaultPipeline
from aixplain.modules import asset

{% for spec in specs %}

class {{ spec.class_name }}Inputs(Inputs):
{% for input in spec.inputs %}
    {{ input.name }}: InputParam = None
{% endfor %}

    def __init__(self, node=None):
        super().__init__(node=node)
{% for input in spec.inputs %}
        self.{{ input.name }} = self.create_param(code="{{ input.name }}", data_type=DataType.{{ input.data_type | upper }}, is_required={{ input.is_required }})
{% endfor %}


class {{ spec.class_name }}Outputs(Outputs):
{% for output in spec.outputs %}
    {{ output.name }}: OutputParam = None
{% endfor %}
{% if spec.is_segmentor %}
    audio: OutputParam = None
{% endif %}

    def __init__(self, node=None):
        super().__init__(node=node)
{% for output in spec.outputs %}
        self.{{ output.name }} = self.create_param(code="{{ output.name }}", data_type=DataType.{{ output.data_type | upper }})
{% endfor %}
{% if spec.is_segmentor %}
        self.audio = self.create_param(code="audio", data_type=DataType.AUDIO)
{% endif %}


class {{ spec.class_name }}({{spec.base_class}}[{{ spec.class_name }}Inputs, {{ spec.class_name }}Outputs]):
    \"\"\"
    {{ spec.description | wordwrap }}

    InputType: {{ spec.input_type }}
    OutputType: {{ spec.output_type }}
    \"\"\"
    function: str = "{{ spec.id }}"
    input_type: str = DataType.{{ spec.input_type | upper }}
    output_type: str = DataType.{{ spec.output_type | upper }}

    inputs_class: Type[TI] = {{ spec.class_name }}Inputs
    outputs_class: Type[TO] = {{ spec.class_name }}Outputs

{% endfor %}


class Pipeline(DefaultPipeline):

{% for spec in specs %}
    def {{ spec.function_name }}(self, asset_id: Union[str, asset.Asset], *args, **kwargs) -> {{ spec.class_name }}:
        \"\"\"
        {{ spec.description | wordwrap }}
        \"\"\"
        return {{ spec.class_name }}(*args, asset_id=asset_id, pipeline=self, **kwargs)

{% endfor %}
"""


def fetch_functions():
    """
    Fetch functions from the backend
    """
    api_key = config.TEAM_API_KEY
    aixplain_key = config.AIXPLAIN_API_KEY
    backend_url = config.BACKEND_URL

    url = urljoin(backend_url, "sdk/functions")
    headers = {
        "Content-Type": "application/json",
    }

    if aixplain_key:
        headers["x-aixplain-key"] = aixplain_key
    else:
        headers["x-api-key"] = api_key

    r = requests.get(url, headers=headers)
    try:
        r.raise_for_status()
    except requests.exceptions.HTTPError as e:
        print("Functions could not be loaded, see error below")
        raise e

    resp = r.json()
    return resp["items"]


def populate_data_types(functions: list):
    """
    Populate the data types
    """
    data_types = set()
    for function in functions:
        for param in function["params"]:
            data_types.add(param["dataType"])
        for output in function["output"]:
            data_types.add(output["dataType"])
    return data_types


def populate_specs(functions: list):
    """
    Populate the function class specs
    """
    function_class_specs = []
    for function in functions:
        # slugify function name by trimming some special chars and
        # transforming it to snake case
        function_name = (
            function["id"]
            .replace("-", "_")
            .replace("(", "_")
            .replace(")", "_")
        )
        base_class = "AssetNode"
        is_segmentor = function["id"] in SEGMENTOR_FUNCTIONS
        is_reconstructor = function["id"] in RECONSTRUCTOR_FUNCTIONS
        if is_segmentor:
            base_class = "BaseSegmentor"
        elif is_reconstructor:
            base_class = "BaseReconstructor"

        spec = {
            "id": function["id"],
            "is_segmentor": function["id"] in SEGMENTOR_FUNCTIONS,
            "is_reconstructor": function["id"] in RECONSTRUCTOR_FUNCTIONS,
            "function_name": function_name,
            "base_class": base_class,
            "class_name": "".join(
                [w.title() for w in function_name.split("_")]
            ),
            "description": function["metaData"]["description"],
            "input_type": function["metaData"]["InputType"],
            "output_type": function["metaData"]["OutputType"],
            "inputs": [
                {
                    "name": param["code"],
                    "data_type": param["dataType"],
                    "is_required": param["required"],
                    "is_list": param.get("multipleValues", False),
                    "default": param.get("defaultValues"),
                    "is_fixed": param.get("isFixed", False),
                }
                for param in function["params"]
            ],
            "outputs": [
                {
                    "name": output["code"],
                    "data_type": output["dataType"],
                    "default": output.get("defaultValue"),
                }
                for output in function["output"]
            ],
        }

        function_class_specs.append(spec)

    return function_class_specs


if __name__ == "__main__":
    print("Fetching function specs")

    functions = fetch_functions()
    data_types = populate_data_types(functions)
    specs = populate_specs(functions)

    print(
        f"Populating module with {len(data_types)} data types and {len(specs)} specs"
    )
    env = Environment(
        loader=BaseLoader(),
        trim_blocks=True,
        lstrip_blocks=True,
    )
    template = env.from_string(TEMPLATE)
    output = template.render(data_types=data_types, specs=specs)

    current_dir = pathlib.Path(__file__).parent
    file_path = current_dir / f"{MODULE_NAME}.py"

    print(f"Writing module to file: {file_path}")
    with open(file_path, "w") as f:
        f.write(output)

    print("Module generated successfully")



__author__ = "thiagocastroferreira"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pytest
import os
import requests
from aixplain.factories import DatasetFactory, PipelineFactory


def test_list_pipelines():
    search_result = PipelineFactory.list()

    assert "results" in search_result
    assert "page_total" in search_result
    assert "page_number" in search_result
    assert "total" in search_result
    assert len(search_result["results"]) > 0


def test_get_pipeline():
    reference_pipeline = PipelineFactory.list()["results"][0]

    hypothesis_pipeline = PipelineFactory.get(reference_pipeline.id)
    assert hypothesis_pipeline.id == reference_pipeline.id


@pytest.mark.parametrize(
    "batchmode,version",
    [
        (True, "2.0"),
        (True, "3.0"),
        (False, "2.0"),
        (False, "3.0"),
    ],
)
def test_run_single_str(batchmode: bool, version: str):
    pipeline = PipelineFactory.list(query="SingleNodePipeline")["results"][0]

    response = pipeline.run(data="Translate this thing", **{"batchmode": batchmode, "version": version})
    assert response["status"] == "SUCCESS"


@pytest.mark.parametrize(
    "batchmode,version",
    [
        (True, "2.0"),
        (True, "3.0"),
        (False, "2.0"),
        (False, "3.0"),
    ],
)
def test_run_single_local_file(batchmode: bool, version: str):
    pipeline = PipelineFactory.list(query="SingleNodePipeline")["results"][0]

    fname = "translate_this.txt"
    with open(fname, "w") as f:
        f.write("Translate this thing")

    response = pipeline.run(data=fname, **{"batchmode": batchmode, "version": version})
    os.remove(fname)
    assert response["status"] == "SUCCESS"


@pytest.mark.parametrize(
    "batchmode,version",
    [
        (True, "2.0"),
        (True, "3.0"),
        (False, "2.0"),
        (False, "3.0"),
    ],
)
def test_run_with_url(batchmode: bool, version: str):
    pipeline = PipelineFactory.list(query="SingleNodePipeline")["results"][0]

    response = pipeline.run(
        data="https://aixplain-platform-assets.s3.amazonaws.com/data/dev/64c81163f8bdcac7443c2dad/data/f8.txt",
        **{"batchmode": batchmode, "version": version}
    )
    assert response["status"] == "SUCCESS"


@pytest.mark.parametrize(
    "batchmode,version",
    [
        (True, "2.0"),
        (True, "3.0"),
        (False, "2.0"),
        (False, "3.0"),
    ],
)
def test_run_with_dataset(batchmode: bool, version: str):
    dataset = DatasetFactory.list(query="for_functional_tests")["results"][0]
    data_asset_id = dataset.id
    data_id = dataset.source_data["en"].id
    pipeline = PipelineFactory.list(query="SingleNodePipeline")["results"][0]

    response = pipeline.run(data=data_id, data_asset=data_asset_id, **{"batchmode": batchmode, "version": version})
    assert response["status"] == "SUCCESS"


@pytest.mark.parametrize(
    "batchmode,version",
    [
        (True, "2.0"),
        (True, "3.0"),
        (False, "2.0"),
        (False, "3.0"),
    ],
)
def test_run_multipipe_with_strings(batchmode: bool, version: str):
    pipeline = PipelineFactory.list(query="MultiInputPipeline")["results"][0]

    response = pipeline.run(
        data={"Input": "Translate this thing.", "Reference": "Traduza esta coisa."},
        **{"batchmode": batchmode, "version": version}
    )
    assert response["status"] == "SUCCESS"


@pytest.mark.parametrize(
    "batchmode,version",
    [
        (True, "2.0"),
        (True, "3.0"),
        (False, "2.0"),
        (False, "3.0"),
    ],
)
def test_run_multipipe_with_datasets(batchmode: bool, version: str):
    pipeline = PipelineFactory.list(query="MultiInputPipeline")["results"][0]

    dataset = DatasetFactory.list(query="for_functional_tests")["results"][0]
    data_asset_id = dataset.id

    input_id = dataset.source_data["en"].id
    reference_id = dataset.target_data["pt"][0].id

    response = pipeline.run(
        data={"Input": input_id, "Reference": reference_id},
        data_asset={"Input": data_asset_id, "Reference": data_asset_id},
        **{"batchmode": batchmode, "version": version}
    )
    assert response["status"] == "SUCCESS"


@pytest.mark.parametrize("version", ["2.0", "3.0"])
def test_run_segment_reconstruct(version: str):
    pipeline = PipelineFactory.list(query="Segmentation/Reconstruction Functional Test - DO NOT DELETE")["results"][0]
    response = pipeline.run("https://aixplain-platform-assets.s3.amazonaws.com/samples/en/CPAC1x2.wav", **{"version": version})

    assert response["status"] == "SUCCESS"
    output = response["data"][0]
    assert output["label"] == "Output 1"


@pytest.mark.parametrize("version", ["2.0", "3.0"])
def test_run_translation_metric(version: str):
    dataset = DatasetFactory.list(query="for_functional_tests")["results"][0]
    data_asset_id = dataset.id

    reference_id = dataset.target_data["pt"][0].id

    pipeline = PipelineFactory.list(query="Translation Metric Functional Test - DO NOT DELETE")["results"][0]
    response = pipeline.run(
        data={"TextInput": reference_id, "ReferenceInput": reference_id},
        data_asset={"TextInput": data_asset_id, "ReferenceInput": data_asset_id},
        **{"version": version}
    )

    assert response["status"] == "SUCCESS"
    data = response["data"][0]["segments"][0]["response"]
    data = requests.get(data).text
    assert float(data) == 100.0


@pytest.mark.parametrize("version", ["2.0", "3.0"])
def test_run_metric(version: str):
    pipeline = PipelineFactory.list(query="ASR Metric Functional Test - DO NOT DELETE")["results"][0]
    response = pipeline.run(
        {
            "AudioInput": "https://aixplain-platform-assets.s3.amazonaws.com/samples/en/CPAC1x2.wav",
            "ReferenceInput": "https://aixplain-platform-assets.s3.amazonaws.com/samples/en/CPAC1x2.txt",
        },
        **{"version": version}
    )

    assert response["status"] == "SUCCESS"
    assert len(response["data"]) == 2
    assert response["data"][0]["label"] in ["TranscriptOutput", "ScoreOutput"]
    assert response["data"][1]["label"] in ["TranscriptOutput", "ScoreOutput"]


@pytest.mark.parametrize(
    "input_data,output_data,version",
    [
        ("https://aixplain-platform-assets.s3.amazonaws.com/samples/en/CPAC1x2.wav", "AudioOutput", "2.0"),
        ("https://aixplain-platform-assets.s3.amazonaws.com/samples/en/CPAC1x2.txt", "TextOutput", "2.0"),
        ("https://aixplain-platform-assets.s3.amazonaws.com/samples/en/CPAC1x2.wav", "AudioOutput", "3.0"),
        ("https://aixplain-platform-assets.s3.amazonaws.com/samples/en/CPAC1x2.txt", "TextOutput", "3.0"),
    ],
)
def test_run_router(input_data: str, output_data: str, version: str):
    pipeline = PipelineFactory.list(query="Router Test - DO NOT DELETE")["results"][0]
    response = pipeline.run(input_data, **{"version": version})

    assert response["status"] == "SUCCESS"
    assert response["data"][0]["label"] == output_data


@pytest.mark.parametrize(
    "input_data,output_data,version",
    [
        ("I love it.", "PositiveOutput", "2.0"),
        ("I hate it.", "NegativeOutput", "2.0"),
        ("I love it.", "PositiveOutput", "3.0"),
        ("I hate it.", "NegativeOutput", "3.0"),
    ],
)
def test_run_decision(input_data: str, output_data: str, version: str):
    pipeline = PipelineFactory.list(query="Decision Test - DO NOT DELETE")["results"][0]
    response = pipeline.run(input_data, **{"version": version})

    assert response["status"] == "SUCCESS"
    assert response["data"][0]["label"] == output_data



__author__ = "aiXplain"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Duraikrishna Selvaraju, Thiago Castro Ferreira, Shreyas Sharma and Lucas Pavanelli
Date: September 1st 2022
Description:
    Pipeline Class
"""

import time
import json
import os
import logging
from aixplain.modules.asset import Asset
from aixplain.utils import config
from aixplain.utils.file_utils import _request_with_retry
from typing import Dict, Optional, Text, Union
from urllib.parse import urljoin


class Pipeline(Asset):
    """Representing a custom pipeline that was created on the aiXplain Platform

    Attributes:
        id (Text): ID of the Pipeline
        name (Text): Name of the Pipeline
        api_key (Text): Team API Key to run the Pipeline.
        url (Text, optional): running URL of platform. Defaults to config.BACKEND_URL.
        supplier (Text, optional): Pipeline supplier. Defaults to "aiXplain".
        version (Text, optional): version of the pipeline. Defaults to "1.0".
        **additional_info: Any additional Pipeline info to be saved
    """

    def __init__(
        self,
        id: Text,
        name: Text,
        api_key: Text,
        url: Text = config.BACKEND_URL,
        supplier: Text = "aiXplain",
        version: Text = "1.0",
        **additional_info,
    ) -> None:
        """Create a Pipeline with the necessary information

        Args:
            id (Text): ID of the Pipeline
            name (Text): Name of the Pipeline
            api_key (Text): Team API Key to run the Pipeline.
            url (Text, optional): running URL of platform. Defaults to config.BACKEND_URL.
            supplier (Text, optional): Pipeline supplier. Defaults to "aiXplain".
            version (Text, optional): version of the pipeline. Defaults to "1.0".
            **additional_info: Any additional Pipeline info to be saved
        """
        if not name:
            raise ValueError("Pipeline name is required")

        super().__init__(id, name, "", supplier, version)
        self.api_key = api_key
        self.url = f"{url}/assets/pipeline/execution/run"
        self.additional_info = additional_info

    def __polling(
        self,
        poll_url: Text,
        name: Text = "pipeline_process",
        wait_time: float = 1.0,
        timeout: float = 20000.0,
    ) -> Dict:
        """Keeps polling the platform to check whether an asynchronous call is done.

        Args:
            poll_url (str): polling URL
            name (str, optional): ID given to a call. Defaults to "pipeline_process".
            wait_time (float, optional): wait time in seconds between polling calls. Defaults to 1.0.
            timeout (float, optional): total polling time. Defaults to 20000.0.

        Returns:
            dict: response obtained by polling call
        """
        # TO DO: wait_time = to the longest path of the pipeline * minimum waiting time
        logging.debug(f"Polling for Pipeline: Start polling for {name} ")
        start, end = time.time(), time.time()
        completed = False
        response_body = {"status": "FAILED"}
        while not completed and (end - start) < timeout:
            try:
                response_body = self.poll(poll_url, name=name)
                logging.debug(f"Polling for Pipeline: Status of polling for {name} : {response_body}")
                completed = response_body["completed"]

                end = time.time()
                if completed is False:
                    time.sleep(wait_time)
                    if wait_time < 60:
                        wait_time *= 1.1
            except Exception:
                logging.error(f"Polling for Pipeline: polling for {name} : Continue")
        if response_body and response_body["status"] == "SUCCESS":
            try:
                logging.debug(f"Polling for Pipeline: Final status of polling for {name} : SUCCESS - {response_body}")
            except Exception:
                logging.error(f"Polling for Pipeline: Final status of polling for {name} : ERROR - {response_body}")
        else:
            logging.error(
                f"Polling for Pipeline: Final status of polling for {name} : No response in {timeout} seconds - {response_body}"
            )
        return response_body

    def poll(self, poll_url: Text, name: Text = "pipeline_process") -> Dict:
        """Poll the platform to check whether an asynchronous call is done.

        Args:
            poll_url (Text): polling URL
            name (Text, optional): ID given to a call. Defaults to "pipeline_process".

        Returns:
            Dict: response obtained by polling call
        """

        headers = {
            "x-api-key": self.api_key,
            "Content-Type": "application/json",
        }
        r = _request_with_retry("get", poll_url, headers=headers)
        try:
            resp = r.json()
            logging.info(f"Single Poll for Pipeline: Status of polling for {name} : {resp}")
        except Exception:
            resp = {"status": "FAILED"}
        return resp

    def run(
        self,
        data: Union[Text, Dict],
        data_asset: Optional[Union[Text, Dict]] = None,
        name: Text = "pipeline_process",
        timeout: float = 20000.0,
        wait_time: float = 1.0,
        **kwargs,
    ) -> Dict:
        """Runs a pipeline call.

        Args:
            data (Union[Text, Dict]): link to the input data
            data_asset (Optional[Union[Text, Dict]], optional): Data asset to be processed by the pipeline. Defaults to None.
            name (Text, optional): ID given to a call. Defaults to "pipeline_process".
            timeout (float, optional): total polling time. Defaults to 20000.0.
            wait_time (float, optional): wait time in seconds between polling calls. Defaults to 1.0.
            kwargs: A dictionary of keyword arguments. The keys are the argument names

        Returns:
            Dict: parsed output from pipeline
        """
        start = time.time()
        try:
            response = self.run_async(data, data_asset=data_asset, name=name, **kwargs)
            if response["status"] == "FAILED":
                end = time.time()
                response["elapsed_time"] = end - start
                return response
            poll_url = response["url"]
            end = time.time()
            response = self.__polling(poll_url, name=name, timeout=timeout, wait_time=wait_time)
            return response
        except Exception as e:
            error_message = f"Error in request for {name}: {str(e)}"
            logging.error(error_message)
            logging.exception(error_message)
            end = time.time()
            return {
                "status": "FAILED",
                "error": error_message,
                "elapsed_time": end - start,
            }

    def __prepare_payload(
        self,
        data: Union[Text, Dict],
        data_asset: Optional[Union[Text, Dict]] = None,
    ) -> Dict:
        """Prepare pipeline execution payload, validating the input data

        Args:
            data (Union[Text, Dict]): input data
            data_asset (Optional[Union[Text, Dict]], optional): input data asset. Defaults to None.

        Returns:
            Dict: pipeline execution payload
        """
        from aixplain.factories import (
            CorpusFactory,
            DatasetFactory,
            FileFactory,
        )

        # if an input data asset is provided, just handle the data
        if data_asset is None:
            # upload the data when a local path is provided
            data = FileFactory.to_link(data)
            if isinstance(data, dict):
                payload = data
                for key in payload:
                    payload[key] = {"value": payload[key]}

                for node_label in payload:
                    payload[node_label]["nodeId"] = node_label

                payload = {"data": list(payload.values())}
            else:
                try:
                    payload = json.loads(data)
                    if isinstance(payload, dict) is False:
                        if isinstance(payload, int) is True or isinstance(payload, float) is True:
                            payload = str(payload)
                        payload = {"data": payload}
                except Exception:
                    payload = {"data": data}
        else:
            payload = {}
            if isinstance(data_asset, str) is True:
                data_asset = {"1": data_asset}

                # make sure data asset and data are provided in the same format,
                # mostly when in a multi-input scenario, where a dictionary should be provided.
                if isinstance(data, dict) is True:
                    raise Exception(
                        'Pipeline Run Error: Similar to "data", please specify the node input label where the data asset should be set in "data_asset".'
                    )
                else:
                    data = {"1": data}
            elif isinstance(data, str) is True:
                raise Exception(
                    'Pipeline Run Error: Similar to "data_asset", please specify the node input label where the data should be set in "data".'
                )

            # validate the existence of data asset and data
            for node_label in data_asset:
                asset_payload = {"dataAsset": {}}
                data_asset_found, data_found = True, False
                try:
                    dasset = CorpusFactory.get(str(data_asset[node_label]))
                    asset_payload["dataAsset"]["corpus_id"] = dasset.id
                    if len([d for d in dasset.data if d.id == data[node_label]]) > 0:
                        data_found = True
                except Exception:
                    try:
                        dasset = DatasetFactory.get(str(data_asset[node_label]))
                        asset_payload["dataAsset"]["dataset_id"] = dasset.id

                        if (
                            len([dfield for dfield in dasset.source_data if dasset.source_data[dfield].id == data[node_label]])
                            > 0
                        ):
                            data_found = True
                        else:
                            for target in dasset.target_data:
                                for target_row in dasset.target_data[target]:
                                    if target_row.id == data[node_label]:
                                        data_found = True
                                        break
                                if data_found is True:
                                    break
                    except Exception:
                        data_asset_found = False
                if data_asset_found is False:
                    raise Exception(
                        f'Pipeline Run Error: Data Asset "{data_asset[node_label]}" not found. Make sure this asset exists or you have access to it.'
                    )
                elif data_found is False:
                    raise Exception(
                        f'Pipeline Run Error: Data "{data[node_label]}" not found in Data Asset "{data_asset[node_label]}" not found.'
                    )

                asset_payload["dataAsset"]["data_id"] = data[node_label]
                payload[node_label] = asset_payload

            if len(payload) > 1:
                for node_label in payload:
                    payload[node_label]["nodeId"] = node_label
            payload = {"data": list(payload.values())}
        return payload

    def run_async(
        self, data: Union[Text, Dict], data_asset: Optional[Union[Text, Dict]] = None, name: Text = "pipeline_process", **kwargs
    ) -> Dict:
        """Runs asynchronously a pipeline call.

        Args:
            data (Union[Text, Dict]): link to the input data
            data_asset (Optional[Union[Text, Dict]], optional): Data asset to be processed by the pipeline. Defaults to None.
            name (Text, optional): ID given to a call. Defaults to "pipeline_process".
            kwargs: A dictionary of keyword arguments. The keys are the argument names

        Returns:
            Dict: polling URL in response
        """
        headers = {
            "x-api-key": self.api_key,
            "Content-Type": "application/json",
        }

        payload = self.__prepare_payload(data=data, data_asset=data_asset)
        payload.update(kwargs)
        payload = json.dumps(payload)
        call_url = f"{self.url}/{self.id}"
        logging.info(f"Start service for {name}  - {call_url} - {payload}")
        r = _request_with_retry("post", call_url, headers=headers, data=payload)

        resp = None
        try:
            if 200 <= r.status_code < 300:
                resp = r.json()
                logging.info(f"Result of request for {name}  - {r.status_code} - {resp}")
                poll_url = resp["url"]
                response = {"status": "IN_PROGRESS", "url": poll_url}
            else:
                if r.status_code == 401:
                    error = "Unauthorized API key: Please verify the spelling of the API key and its current validity."
                elif 460 <= r.status_code < 470:
                    error = "Subscription-related error: Please ensure that your subscription is active and has not expired."
                elif 470 <= r.status_code < 480:
                    error = "Billing-related error: Please ensure you have enough credits to run this pipeline. "
                elif 480 <= r.status_code < 490:
                    error = "Supplier-related error: Please ensure that the selected supplier provides the pipeline you are trying to access."
                elif 490 <= r.status_code < 500:
                    error = "Validation-related error: Please ensure all required fields are provided and correctly formatted."
                else:
                    status_code = str(r.status_code)
                    error = f"Status {status_code}: Unspecified error: An unspecified error occurred while processing your request."
                response = {"status": "FAILED", "error_message": error}
                logging.error(f"Error in request for {name} - {r.status_code}: {error}")               
        except Exception:
            response = {"status": "FAILED"}
            if resp is not None:
                response["error"] = resp
        return response

    def update(
        self,
        pipeline: Union[Text, Dict],
        save_as_asset: bool = False,
        api_key: Optional[Text] = None,
    ):
        """Update Pipeline

        Args:
            pipeline (Union[Text, Dict]): Pipeline as a Python dictionary or in a JSON file
            save_as_asset (bool, optional): Save as asset (True) or draft (False). Defaults to False.
            api_key (Optional[Text], optional): Team API Key to create the Pipeline. Defaults to None.

        Raises:
            Exception: Make sure the pipeline to be save is in a JSON file.
        """
        try:
            if isinstance(pipeline, str) is True:
                _, ext = os.path.splitext(pipeline)
                assert (
                    os.path.exists(pipeline) and ext == ".json"
                ), "Pipeline Update Error: Make sure the pipeline to be saved is in a JSON file."
                with open(pipeline) as f:
                    pipeline = json.load(f)

            for i, node in enumerate(pipeline["nodes"]):
                if "functionType" in node and node["functionType"] == "AI":
                    pipeline["nodes"][i]["functionType"] = pipeline["nodes"][i]["functionType"].lower()
            # prepare payload
            status = "draft"
            if save_as_asset is True:
                status = "onboarded"
            payload = {
                "name": self.name,
                "status": status,
                "architecture": pipeline,
            }
            url = urljoin(config.BACKEND_URL, f"sdk/pipelines/{self.id}")
            api_key = api_key if api_key is not None else config.TEAM_API_KEY
            headers = {
                "Authorization": f"Token {api_key}",
                "Content-Type": "application/json",
            }
            logging.info(f"Start service for PUT Update Pipeline - {url} - {headers} - {json.dumps(payload)}")
            r = _request_with_retry("put", url, headers=headers, json=payload)
            response = r.json()
            logging.info(f"Pipeline {response['id']} Updated.")
        except Exception as e:
            raise Exception(e)

    def delete(self) -> None:
        """Delete Dataset service"""
        try:
            url = urljoin(config.BACKEND_URL, f"sdk/pipelines/{self.id}")
            headers = {
                "Authorization": f"Token {config.TEAM_API_KEY}",
                "Content-Type": "application/json",
            }
            logging.info(f"Start service for DELETE Pipeline  - {url} - {headers}")
            r = _request_with_retry("delete", url, headers=headers)
            if r.status_code != 200:
                raise Exception()
        except Exception:
            message = "Pipeline Deletion Error: Make sure the pipeline exists and you are the owner."
            logging.error(message)
            raise Exception(f"{message}")

    def save(self, save_as_asset: bool = False, api_key: Optional[Text] = None):
        """Save Pipeline

        Args:
            save_as_asset (bool, optional): Save as asset (True) or draft (False). Defaults to False.
            api_key (Optional[Text], optional): Team API Key to create the Pipeline. Defaults to None.

        Raises:
            Exception: Make sure the pipeline to be save is in a JSON file.
        """
        try:
            pipeline = self.to_dict()

            for i, node in enumerate(pipeline["nodes"]):
                if "functionType" in node and node["functionType"] == "AI":
                    pipeline["nodes"][i]["functionType"] = pipeline["nodes"][i]["functionType"].lower()
            # prepare payload
            status = "draft"
            if save_as_asset is True:
                status = "onboarded"
            payload = {
                "name": self.name,
                "status": status,
                "architecture": pipeline,
            }

            if self.id != "":
                method = "put"
                url = urljoin(config.BACKEND_URL, f"sdk/pipelines/{self.id}")
            else:
                method = "post"
                url = urljoin(config.BACKEND_URL, "sdk/pipelines")
            api_key = api_key if api_key is not None else config.TEAM_API_KEY
            headers = {
                "Authorization": f"Token {api_key}",
                "Content-Type": "application/json",
            }
            logging.info(f"Start service for Save Pipeline - {url} - {headers} - {json.dumps(payload)}")
            r = _request_with_retry(method, url, headers=headers, json=payload)
            response = r.json()
            self.id = response["id"]
            logging.info(f"Pipeline {response['id']} Saved.")
        except Exception as e:
            raise Exception(e)



__author__ = "thiagocastroferreira"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pytest
import time

from uuid import uuid4
from aixplain.enums import Function, Language, License, Privacy, DataSubtype, DataType, StorageType, OnboardStatus
from aixplain.factories import DatasetFactory
from aixplain.modules import MetaData


@pytest.fixture
def meta1():
    return [
        {
            "name": "audio",
            "dtype": "audio",
            "storage_type": "url",
            "start_column": "audio_start_time",
            "end_column": "audio_end_time",
            "languages": [Language.English_UNITED_STATES],
        }
    ]


@pytest.fixture
def meta2():
    return [{"name": "text", "dtype": "text", "storage_type": "text", "languages": [Language.English_UNITED_STATES]}]


@pytest.fixture
def split():
    return MetaData(
        name="split",
        dtype=DataType.LABEL,
        dsubtype=DataSubtype.SPLIT,
        storage_type=StorageType.TEXT,
    )


def test_dataset_onboard_get_delete(meta1, meta2):
    upload_file = "tests/functional/data_asset/input/audio-en_url.csv"

    response = DatasetFactory.create(
        name=str(uuid4()),
        description="Test dataset",
        license=License.MIT,
        function=Function.SPEECH_RECOGNITION,
        content_path=upload_file,
        input_schema=meta1,
        output_schema=meta2,
        tags=[],
        privacy=Privacy.PRIVATE,
    )
    asset_id = response["asset_id"]
    onboard_status = OnboardStatus(response["status"])
    while onboard_status == OnboardStatus.ONBOARDING:
        dataset = DatasetFactory.get(asset_id)
        onboard_status = dataset.onboard_status
        time.sleep(1)
    # assert the asset was onboarded
    assert onboard_status == OnboardStatus.ONBOARDED
    # assert the asset was deleted
    dataset.delete()
    with pytest.raises(Exception):
        dataset = DatasetFactory.get(asset_id)


def test_invalid_dataset_onboard(meta1, meta2):
    upload_file = "tests/functional/data_asset/input/audio-en_url.csv"

    with pytest.raises(Exception):
        response = DatasetFactory.create(
            name=str(uuid4()),
            description="Test dataset",
            license=License.MIT,
            function=Function.TRANSLATION,
            content_path=upload_file,
            input_schema=meta1,
            output_schema=meta2,
            tags=[],
            privacy=Privacy.PRIVATE,
        )


def test_dataset_listing():
    response = DatasetFactory.list()
    assert "results" in response


def test_dataset_get_error():
    with pytest.raises(Exception):
        response = DatasetFactory.get("131312")


def test_invalid_dataset_splitting(meta1, meta2, split):
    upload_file = "tests/functional/data_asset/input/audio-en_with_invalid_split_url.csv"

    split2 = MetaData(
        name="split-2",
        dtype=DataType.LABEL,
        dsubtype=DataSubtype.SPLIT,
        storage_type=StorageType.TEXT,
    )

    with pytest.raises(Exception):
        response = DatasetFactory.create(
            name=str(uuid4()),
            description="Test dataset",
            license=License.MIT,
            function=Function.SPEECH_RECOGNITION,
            content_path=upload_file,
            input_schema=meta1,
            output_schema=meta2,
            metadata_schema=[split, split2],
            tags=[],
            privacy=Privacy.PRIVATE,
        )


def test_valid_dataset_splitting(meta1, meta2, split):
    upload_file = "tests/functional/data_asset/input/audio-en_with_split_url.csv"

    response = DatasetFactory.create(
        name=str(uuid4()),
        description="Test dataset",
        license=License.MIT,
        function=Function.SPEECH_RECOGNITION,
        content_path=upload_file,
        input_schema=meta1,
        output_schema=meta2,
        metadata_schema=[split],
        tags=[],
        privacy=Privacy.PRIVATE,
    )

    assert response["status"] == "onboarding"
    asset_id = response["asset_id"]
    onboard_status = OnboardStatus(response["status"])
    while onboard_status == OnboardStatus.ONBOARDING:
        dataset = DatasetFactory.get(asset_id)
        onboard_status = dataset.onboard_status
        time.sleep(1)
    # assert the asset was onboarded
    assert onboard_status == OnboardStatus.ONBOARDED
    dataset.delete()



import json
from dotenv import load_dotenv

load_dotenv()
from aixplain.factories import ModelFactory, DatasetFactory, MetricFactory, PipelineFactory
from aixplain.modules import LLM
from pathlib import Path
from aixplain.enums import Function, Language, OwnershipType, Supplier, SortBy, SortOrder

import pytest

INPUTS_PATH = Path(r"tests/functional/general_assets/data/asset_run_test_data.json")


@pytest.fixture
def inputs():
    with open(INPUTS_PATH) as f:
        return json.load(f)


def __get_asset_factory(asset_name):
    if asset_name == "model":
        AssetFactory = ModelFactory
    elif asset_name == "dataset":
        AssetFactory = DatasetFactory
    elif asset_name == "metric":
        AssetFactory = MetricFactory
    elif asset_name == "pipeline":
        AssetFactory = PipelineFactory
    return AssetFactory


@pytest.mark.parametrize("asset_name", ["model", "dataset", "metric"])
def test_list(asset_name):
    AssetFactory = __get_asset_factory(asset_name)
    asset_list = AssetFactory.list()
    assert asset_list["page_total"] == len(asset_list["results"])


@pytest.mark.parametrize("asset_name", ["model", "pipeline", "metric"])
def test_run(inputs, asset_name):
    asset_details = inputs[asset_name]
    AssetFactory = __get_asset_factory(asset_name)
    if asset_name == "pipeline":
        asset = AssetFactory.list(query=asset_details["name"])["results"][0]
    else:
        asset = AssetFactory.get(asset_details["id"])
    payload = asset_details["data"]
    if type(payload) is dict:
        output = asset.run(**payload)
    else:
        output = asset.run(payload)
    assert output["completed"] and output["status"] == "SUCCESS"


def test_model_function():
    desired_function = Function.TRANSLATION
    models = ModelFactory.list(function=desired_function)["results"]
    for model in models:
        assert model.function == desired_function


def test_model_supplier():
    desired_suppliers = [Supplier.GOOGLE]
    models = ModelFactory.list(suppliers=desired_suppliers)["results"]
    for model in models:
        assert model.supplier.value in [desired_supplier.value for desired_supplier in desired_suppliers]


def test_model_sort():
    function = Function.TRANSLATION
    src_language = Language.Portuguese
    trg_language = Language.English

    models = ModelFactory.list(
        function=function,
        source_languages=src_language,
        target_languages=trg_language,
        sort_by=SortBy.PRICE,
        sort_order=SortOrder.DESCENDING,
    )["results"]
    for idx in range(1, len(models)):
        prev_model = models[idx - 1]
        model = models[idx]

        prev_model_price = prev_model.cost["price"]
        model_price = model.cost["price"]
        assert prev_model_price >= model_price


def test_model_ownership():
    models = ModelFactory.list(ownership=OwnershipType.SUBSCRIBED)["results"]
    for model in models:
        assert model.is_subscribed is True


def test_model_query():
    query = "Mongo"
    models = ModelFactory.list(query=query)["results"]
    for model in models:
        assert query in model.name


def test_model_deletion():
    """Test that a model cannot be deleted."""
    model = ModelFactory.get("640b517694bf816d35a59125")
    with pytest.raises(Exception):
        model.delete()


def test_llm_instantiation():
    """Test that the LLM model is correctly instantiated."""
    models = ModelFactory.list(function=Function.TEXT_GENERATION)["results"]
    assert isinstance(models[0], LLM)


def test_model_io():
    model_id = "64aee5824d34b1221e70ac07"
    model = ModelFactory.get(model_id)

    expected_input = {
        "text": {
            "name": "Text Prompt",
            "code": "text",
            "required": True,
            "isFixed": False,
            "dataType": "text",
            "dataSubType": "text",
            "multipleValues": False,
            "defaultValues": [],
        }
    }
    expected_output = {"data": {"name": "Generated Image", "code": "data", "defaultValue": [], "dataType": "image"}}

    assert model.input_params == expected_input
    assert model.output_params == expected_output



from .asset import Pipeline as PipelineAsset
from .designer import DesignerPipeline


class DefaultPipeline(PipelineAsset, DesignerPipeline):

    def __init__(self, *args, **kwargs):
        PipelineAsset.__init__(self, *args, **kwargs)
        DesignerPipeline.__init__(self)

    def save(self, *args, **kwargs):
        self.auto_infer()
        self.validate()
        super().save(*args, **kwargs)

    def to_dict(self) -> dict:
        return self.serialize()



__author__ = "thiagocastroferreira"

"""
Copyright 2022 The aiXplain SDK authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from dotenv import load_dotenv
import pytest

load_dotenv()
import requests_mock
from aixplain.utils import config
from aixplain.factories import PipelineFactory
from aixplain.modules import Pipeline
from urllib.parse import urljoin


def test_create_pipeline():
    with requests_mock.Mocker() as mock:
        url = urljoin(config.BACKEND_URL, "sdk/pipelines")
        headers = {"x-api-key": config.TEAM_API_KEY, "Content-Type": "application/json"}
        ref_response = {"id": "12345"}
        mock.post(url, headers=headers, json=ref_response)
        ref_pipeline = Pipeline(id="12345", name="Pipeline Test", api_key=config.TEAM_API_KEY)
        hyp_pipeline = PipelineFactory.create(pipeline={"nodes": []}, name="Pipeline Test")
    assert hyp_pipeline.id == ref_pipeline.id
    assert hyp_pipeline.name == ref_pipeline.name

@pytest.mark.parametrize(
    "status_code,error_message",
    [
        (401,"Unauthorized API key: Please verify the spelling of the API key and its current validity."),
        (465,"Subscription-related error: Please ensure that your subscription is active and has not expired."),
        (475,"Billing-related error: Please ensure you have enough credits to run this pipeline. "),
        (485, "Supplier-related error: Please ensure that the selected supplier provides the pipeline you are trying to access."),
        (495, "Validation-related error: Please ensure all required fields are provided and correctly formatted."),
        (501, "Status 501: Unspecified error: An unspecified error occurred while processing your request."),

    ],
)

def test_run_async_errors(status_code, error_message):
    base_url = config.BACKEND_URL
    pipeline_id = "pipeline_id"
    execute_url = f"{base_url}/assets/pipeline/execution/run/{pipeline_id}"
    
    with requests_mock.Mocker() as mock:
        mock.post(execute_url, status_code=status_code)
        test_pipeline = Pipeline(id=pipeline_id, api_key=config.TEAM_API_KEY, name="Test Pipeline", url=base_url)
        response = test_pipeline.run_async(data="input_data")
    assert response["status"] == "FAILED"
    assert response["error_message"] == error_message